/*

  Elmo driver for P0 and later hardware

  Copyright TiVo Inc., 2003-2004.   

*/

#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/mman.h>
#include <linux/linkage.h>
#include <linux/fs.h>
#include <linux/time.h>
#include <linux/wait.h>
#include <linux/blkdev.h>
#include <linux/blk.h>
#include <linux/spinlock.h>
#include <linux/tivoconfig.h>
#include <linux/list.h>
#include <linux/hdreg.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>

#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <asm/semaphore.h>
#include <asm/bcm7315.h>

#include "elmointerface.h"
#include "elmo.h"

#include "ElmoPlog.h"

#if (NDEBUG & NDEBUG_LISTS)
#define LIST(x,y) {printk("LINE:%d   Adding %p to %p\n", __LINE__, (void*)(x), (void*)(y)); if ((x)==(y)) udelay(5); }
#define REMOVE(w,x,y,z) {printk("LINE:%d   Removing: %p->%p  %p->%p \n", __LINE__, (void*)(w), (void*)(x), (void*)(y), (void*)(z)); if ((x)==(y)) udelay(5); }
#else
#define LIST(x,y)
#define REMOVE(w,x,y,z)
#endif

#define assert(xpr)                                     \
  do {                                                  \
    if (!(xpr)) {                                       \
      printk("elmo: ASSERTION FAILURE: %s at %s:%d\n",  \
             #xpr, __FILE__, __LINE__);                \
    }                                                   \
  } while(0)


#define ELMO_DEBUG KERN_DEBUG

#define USLEEP_SLEEP (20*HZ/1000)
#define USLEEP_WAITLONG USLEEP_SLEEP

#undef SIMULATE_IRQ
#undef DEBUG_IOVEC
//#define DEBUG_IOVEC 1

#define ELMO_ABORT_PENDING              -1
#define ELMO_ABORT_NOT_RUNNING          -2
#define ELMO_RESET_SUCCESS              -3

const static char version[] = "2.0";

/*
  Definitions for dealing with the 7315/7317
*/

#define PHYSADDR(base,offset) ((base) + (offset))

/*
  bit 4 set to 0 selects GPIO 33
  bit 5 set to 0 selects GPIO 3
*/
#define BCM_GPIO_SEL0        (PHYSADDR(BCM7315_REG_BASE,0x8007))
#define BCM_GPIO_SEL1        (PHYSADDR(BCM7315_REG_BASE,0x8006))
#define BCM_SPI_EBI_REQ_MUX  (PHYSADDR(BCM7315_REG_BASE,0x80A0))

/*
  PLOAD:
  GPIO_BASE + 0x3 bit 3 is GPIO3 data bit
  GPIO_BASE + 0x7 bit 3 is GPIO3 direction bit
  GPIO_BASE + 0x8 bit 3 is GPIO3 open drain bit

  PROGRAM:
  GPIO_BASE + 0x51 bit 1 is GPIO33 data bit
  GPIO_BASE + 0x53 bit 1 is GPIO33 direction bit
  GPIO_BASE + 0x55 bit 1 is GPIO33 open drain bit
*/

#define BCM_GPIO_BASE     (PHYSADDR(BCM7315_REG_BASE,0x50))

#define CHIP_ID      *((volatile unsigned long *)(PHYSADDR(BCM7315_REG_BASE, 0x0600)))

#define BCM_MTN_BASE            0xffe00000
#define BCM_MTN_CLIENT_INDEX    (volatile unsigned long *)(PHYSADDR(BCM_MTN_BASE, MTN_RB_MMC_CLIENT_INDEX))
#define BCM_MTN_INDEX_DATA_WR   (volatile unsigned long *)(PHYSADDR(BCM_MTN_BASE, MTN_RB_MMC_INDX_DATA_WR))
#define BCM_ISB_MMC_CLIENT_ID   0x12

#define TEI_BASE              ELMO_BRIDGE_BASE
#define TEI_CODE_LOAD         (PHYSADDR(TEI_BASE,0))
#define TEI_REVISION          (PHYSADDR(TEI_BASE,0x1C))

#define ESS_ATA_TRANSFER_ABORT  0xF0

static int elmo_proc_info(char *buffer, char **start, off_t offset, int length, int *eof, void *data);

static void elmo_intr(int irq, void *dev_id, struct pt_regs *regs);
static void do_elmo_intr(int irq, void *dev_id, struct pt_regs *regs);
static void broadcom_intr(int irq, int *useFlag, unsigned long softInterrupt);
static void broadcom_isb_intr(int irq, void *dev_id, struct pt_regs *regs);

static void elmo_wait (int mscount);

static int elmo_check_host_dma_rules (struct elmo_request * cmd);
static int elmo_check_more_host_dma (struct elmo_request * cmd);
#if 0
static int elmo_transfer_to_user (struct elmo_request * cmd, unsigned char *buffer, unsigned int length);
static int elmo_transfer_from_user (struct elmo_request * cmd, unsigned char *buffer, unsigned int length);
#endif /* 0 */

static void run_main(void);
static void elmo_main(void);
static int elmo_probe_bridge(void);
static void elmo_bridge_is_alive();
static void elmo_init_broadcom(void);
static void initialize_SCp(struct elmo_request * request);
static int elmo_run_workers();

static unsigned long elmo_interrupt_for_workers(unsigned long interruptsFired);

static void elmo_queue_worker_for_interrupt (struct elmo_worker *worker,
					   unsigned long interruptsNeeded,
					   enum elmo_worker_state nextState);

static void elmo_queue_worker_for_timer (struct elmo_worker *worker,
				       int timerTicks,
				       enum elmo_worker_state nextState);

static void elmo_worker_next_state (struct elmo_worker *worker,
			      enum elmo_worker_state nextState);

static void elmo_worker_disarm_timer (struct elmo_worker *worker);

static void elmo_worker_set_deadman (struct elmo_worker *worker,
				   unsigned long timerTicks);

static void elmo_worker_disarm_deadman (struct elmo_worker *worker);

static struct elmo_worker* elmo_bulk_io_init(struct elmo_request * request); 

static void elmo_worker_setup(struct elmo_worker *worker);
static void elmo_worker_get_bulk_interface(struct elmo_worker *worker);
static void elmo_worker_get_ess_command(struct elmo_worker *worker);
static void elmo_worker_setup_dma(struct elmo_worker *worker);
static void elmo_worker_await_fifo(struct elmo_worker *worker);
static void elmo_worker_finish_dma(struct elmo_worker *worker);
static void elmo_worker_done(struct elmo_worker *worker);

static void elmo_worker_service_read_fifo(struct elmo_worker *worker);
static void elmo_worker_service_write_fifo(struct elmo_worker *worker);
static void elmo_worker_gate_writer(struct elmo_worker *worker);
static void elmo_worker_become_writer(struct elmo_worker *worker);
static void elmo_worker_writer(struct elmo_worker *worker);
static void elmo_worker_become_reader(struct elmo_worker *worker);
static void elmo_worker_reader(struct elmo_worker *worker);

static void elmo_release_chip_resources (unsigned int resourceMask);

static void elmo_bug(struct elmo_worker *worker, 
		     const char *reason);

static void elmo_task (void *whatever);
static void elmo_timer_task (void *whatever);

static void show_elmo_intr(unsigned long);
static void unpin_memory(Sgat_dio_request *dio_request);
static void show_elmo_errors(elmo_error_codes_e errors);
static void elmo_dump_bridge_regs();

int elmo_ll_open (struct inode *, struct file *);
ssize_t elmo_ll_read (struct file *, char *, size_t, loff_t *);
ssize_t elmo_ll_readv (struct file *filp, const struct iovec *iov, 
		       unsigned long count, loff_t *ppos);
ssize_t elmo_ll_write_common (struct file *, const char *, size_t, loff_t *,
			      int defer);
ssize_t elmo_ll_write (struct file *, const char *, size_t, loff_t *);
ssize_t elmo_ll_write_deferred (struct file *, const char *, size_t, loff_t *);
ssize_t elmo_ll_writev_common (struct file *filp, const struct iovec *iov, 
			       unsigned long count, loff_t *ppos, int defer);
ssize_t elmo_ll_writev (struct file *filp, const struct iovec *iov, 
			unsigned long count, loff_t *ppos);
ssize_t elmo_ll_writev_deferred (struct file *filp, const struct iovec *iov, 
				 unsigned long count, loff_t *ppos);
int elmo_ll_ioctl (struct inode *i, struct file *f, unsigned int cd,
		   unsigned long param);

#define ELMO_MAX_PENDING_WRITERS 5

#define ELMO_BULK_FIFO_SIZE 512

struct BroadcomDMADescriptor {
  int reserved                :  4;
  int transferByteCount       : 12;
  int ob                      :  1;
  int pd                      :  3;
  int l                       :  1;
  int f                       :  1;
  int w                       :  1;
  int pd1                     :  9;
  int memoryAddr              : 32;
};

struct elmo_hostdata {
  volatile unsigned aborted:1;		/* flag, says aborted */
  volatile unsigned pldLoaded:1;        /* flag, bridge is up */
  volatile unsigned initDone:1;         /* flag, initialization done */
  volatile unsigned atapiOK:1;          /* flag, ATAPI usable */
  volatile unsigned wasReset:1;         /* flag, ATAPI bus was reset, need to settle... */
  volatile unsigned mustReset:1;        /* flag, ATAPI reset is required */
  int flags;
  rwlock_t queueLock;                   /* lock on the issue queue */
  unsigned long time_expires;		/* in jiffies, set prior to sleeping */
  int workerSN;
  struct list_head workers;
  volatile unsigned long resourcesBusy; /* On-chip resources */
  unsigned int atapiSectorSize;
  unsigned int atapiBufferModulus;
  int diagMode;
  struct BroadcomDMADescriptor *dmaDescriptors;
  unsigned char *llReadBuffer;
  unsigned int llReadBufferHead, llReadBufferTail;
  unsigned char *llWriteBuffer;
  unsigned int llWriteBufferHead, llWriteBufferTail;
  struct semaphore writeSem;
  unsigned int bulkFifoSize;
  int haveCommand;
  unsigned short ataCommandReg;
  unsigned short ataCylinderReg;
#ifdef ELMO_STATS
  unsigned timebase;			/* Base for time calcs */
  long time_read[8];			/* time to do reads */
  long time_write[8];			/* time to do writes */
  unsigned long bytes_read[8];		/* bytes read */
  unsigned long bytes_write[8];		/* bytes written */
  unsigned pendingr;
  unsigned pendingw;
#endif
  unsigned essWedged;
  volatile unsigned bulkIOCompleted;
  Sgat_dio_request *user_request;
};


static struct file_operations elmo_ll_fops = {
  read:    elmo_ll_read,
  write:   elmo_ll_write,
  ioctl:   elmo_ll_ioctl,
  open:    elmo_ll_open,
  readv:   elmo_ll_readv,
  writev:  elmo_ll_writev,
};

static struct file_operations elmo_ll_deferred_fops = {
  read:   elmo_ll_read,
  write:  elmo_ll_write_deferred,
  ioctl:  elmo_ll_ioctl,
  open:   elmo_ll_open,
  readv:  elmo_ll_readv,
  writev: elmo_ll_writev_deferred,
};

#undef TEMPORARY_SLOW_DMA_HACK

static struct elmo_hostdata elmo_host_data;
static struct proc_dir_entry *elmo_proc_entry;

static void elmo_reset_ATA();

#ifdef SIMULATE_IRQ
static int runIRQs = 0;
#endif

static int elmo_broadcom_isb_enabled = 0;
static int elmo_asic_read_dma_enabled = 0;
static int elmo_asic_write_dma_enabled = 0;

static struct tq_struct elmo_task_struct;
static struct tq_struct elmo_timer_task_struct;

static wait_queue_head_t readerWaitQueue;
static wait_queue_head_t bulkIoWaitQueue;

#if defined(MODULE)
MODULE_AUTHOR("TiVo Inc.");
MODULE_DESCRIPTION("Elmo device driver");
MODULE_LICENSE("GPL");

MODULE_PARM(debug, "1-1i");
MODULE_PARM_DESC(debug,"enable debug loelmo");
static int debug = 0;


MODULE_PARM(swaponread, "0-1i");
MODULE_PARM_DESC(swaponread,"enable byte swapping in FPGA for bulk data read");
static int swaponread=1;

MODULE_PARM(swaponwrite, "0-1i");
MODULE_PARM_DESC(swaponwrite,"enable byte swapping in FPGA for bulk data write");
static int swaponwrite=1;

MODULE_PARM(useshortword, "0-1i");
MODULE_PARM_DESC(useshortword,"use short word (2byte) transfer on low latency path");
static int useshortword=0;

MODULE_PARM(noesswatchdog, "0-1i");
MODULE_PARM_DESC(noesswatchdog,"don't panic on stuck FIFO if noesswatchdog=1");
static int noesswatchdog=0;

static int autodebugExpires = 0;
static int autodebugRestore = -1;

#endif

static const char *bug_line1[4] =
{ "Bit-decay spreads through the Internet",
  "Data inconsistencies disrupt all nearby computers",
  "The home network crashes",
  "Video recordings are lost"};

static const char *bug_line2[4] =
{ "The global economy collapses",
  "PVRs stop working all across America",
  "No one knows who the Survivor is",
  "Riots ensue"};

static int notifier_installed = 0;
#define IS_7317A0()     (((CHIP_ID & 0xFFFFU) == 0) ? 1 : 0)
#define TO_USEC(tv) (((tv.tv_sec * 1000000) + tv.tv_usec))

#ifdef DUMMY_ZERO_COPY_IO
/* DUMMY FUNCTIONS REMOVE THESE : SRIKAR*/

typedef struct sgat_dio_request  /* holding area for scsi scatter gather info */
{
    int timeout;
    unsigned short sclp_count;    /* Count of kernel scatter-gather pieces */
    int dxfer_dir;        /* [i] data transfer direction  */
    struct scatterlist *sclp;
    unsigned short xferId;	
    int      status;
    unsigned long      client_data;
    void *schp; 
    Sgat_dio_hdr_t *hdr;
}Sgat_dio_request;

static int sgat_dio_build_sg_list(const Sgat_dio_hdr_t *hdr, 
                                  Sgat_dio_request *srp)
{
    int i = 0;
    srp->hdr = (Sgat_dio_hdr_t *)hdr;
    srp->dxfer_dir = hdr->dxfer_dir;
    srp->xferId = hdr->xferId;
    srp->sclp = kmalloc(sizeof(struct scatterlist)*hdr->iovec_count, GFP_KERNEL);
    if (!srp->sclp) {
        return -ENOMEM;
    }

    memset(srp->sclp, 0, sizeof(struct scatterlist)*hdr->iovec_count);
    struct scatterlist *sgent = srp->sclp;

    for (i = 0; i < hdr->iovec_count; i++) {
        unsigned char *ptr = kmalloc(hdr->sglp[i].iov_len, GFP_KERNEL);
        if (!ptr) {
            kfree(srp->sclp);
            srp->sclp = NULL;
            return -ENOMEM;
        }

        if (SGAT_DIO_DXFER_TO_DEV == hdr->dxfer_dir) {
            copy_from_user(ptr, hdr->sglp[i].iov_base, hdr->sglp[i].iov_len);
        }
        sgent->address = ptr;
        sgent->length  = hdr->sglp[i].iov_len;
        sgent++;
    }
    srp->sclp_count = hdr->iovec_count;


    return 0;

}

static void sgat_dio_done( Sgat_dio_request *srp)
{
    Sgat_dio_hdr_t *hdr = srp->hdr;
    if (srp->sclp_count != hdr->iovec_count) {
        printk(KERN_ERR "%s : bad iov_count (%d) & sg_count (%d)\n", hdr->iovec_count, srp->sclp_count);
        panic("elmo : bad iovec length in request and header \n");
    }
    struct scatterlist *sgent = srp->sclp;
    int i = 0;
    for (i = 0; i < hdr->iovec_count; i++) {
        if (SGAT_DIO_DXFER_FROM_DEV == hdr->dxfer_dir) {
            copy_to_user(hdr->sglp[i].iov_base, sgent->address, hdr->sglp[i].iov_len);
        }
        kfree(sgent->address);
        sgent++;
    }
    kfree(srp->sclp);
    srp->sclp = NULL;
    hdr->status = srp->status;
    hdr->client_data = srp->client_data;
}
static void sgat_dio_done_nocopy( Sgat_dio_request *srp)
{
    sgat_dio_done(srp);
}

static Sgat_dio_request *alloc_sgat_dio_request(void)
{
    Sgat_dio_request * ptr = kmalloc(sizeof(Sgat_dio_request), GFP_KERNEL);
    if (NULL == ptr) {
        return NULL;
    }
    memset(ptr, 0, sizeof(Sgat_dio_request));
    return ptr;
}

static void free_sgat_dio_request(Sgat_dio_request *srp)
{
    kfree(srp);
    srp = NULL;
}



#endif /* DUMMY_ZERO_COPY_IO */


static void elmo_autodebug_off(void) {
  if (autodebugExpires > 0 && autodebugExpires < jiffies) {
    debug = autodebugRestore;
    autodebugRestore = -1;
    autodebugExpires = 0;
    printk(KERN_INFO "elmo: autodebug mode expired\n");
  }
}

static void elmo_wait( int  mscount )
{
    int delay = ( mscount * HZ ) / 1000;

    wait_queue_head_t wait;

    init_waitqueue_head( &wait );

    if( delay <= 0 ) delay = 1;

    interruptible_sleep_on_timeout( &wait, delay );
}

static unsigned long
ReadRegister( unsigned long addr, unsigned width )
{
    unsigned long value;
    void  *vmem = (void *) (addr);

    switch( width ) {
        case 1:
            value = *((volatile unsigned char*)vmem);
            break;
        case 2:
            value = *((volatile unsigned short*)vmem);
            break;
        case 4:
            value = *((volatile unsigned long*)vmem);
            break;
        default:
            printk( KERN_ERR  "elmo: bad register read width\n" );
	    value = 0;
	    break;
    }

    return value;
}

static void
WriteRegister( unsigned long addr, unsigned value, unsigned width )
{
    void *vmem =(void *) (addr);

    switch( width ) {
        case 1:
            *((volatile unsigned char*)vmem) = value;
            break;
        case 2:
            *((volatile unsigned short*)vmem) = value;
            break;
        case 4:
            *((volatile unsigned long*)vmem) = value;
            break;
        default:
            printk( KERN_ERR  "elmo: bad register write width\n" );
    }
    barrier();
}


#ifdef NOTDEF
static void dump_bridge(void)
{
  int i;
  unsigned short value;
  printk(KERN_INFO "Bridge registers:");
  for (i = 0 ; i <= 32; i += 2) {
    value = *(volatile unsigned short *) KSEG1ADDR(ELMO_BRIDGE_BASE+i);
    printk(" %04X", value);
  }
  printk("\n");
}
#endif

static inline void unpin_memory(Sgat_dio_request *dio_request)
{
      if (NULL != dio_request) {
          sgat_dio_done (dio_request);
          free_sgat_dio_request(dio_request);
      }
}


static void __inline__ elmo_worker_next_state (struct elmo_worker *worker,
					 enum elmo_worker_state nextState) {
      worker->state = nextState;
      worker->go = 1;
      worker->dwell = 0;
}


/**
 *	elmo_init	-	initialise the Elmo adapter isntance
 *
 *	Notes : I assume that the host, hostno, and id bits have been
 *      set correctly.  I don't care about the irq and other fields. 
 *
 *	Locks: interrupts must be enabled when we are called 
 *
 *      XXX Pretty much all of this needs to be redone for Elmo!
 */

static void __init elmo_init()
{
  struct elmo_worker *worker1, *worker2, *worker3;
  unsigned long flags;

  if(in_interrupt())
    printk(KERN_ERR "elmo_init called with interrupts off!\n");

  printk(KERN_INFO "elmo: initializing module\n");
  
  elmo_host_data.aborted = 0;
  elmo_host_data.pldLoaded = elmo_host_data.atapiOK = 0;

  elmo_host_data.bulkIOCompleted = 0;
  elmo_host_data.user_request    = NULL;
  elmo_host_data.essWedged = 0;
  

  elmo_host_data.atapiSectorSize = 2048;
  elmo_host_data.atapiBufferModulus = 2048;
  elmo_host_data.time_expires = 0;
  elmo_host_data.resourcesBusy = 0;
  elmo_host_data.queueLock = RW_LOCK_UNLOCKED;
  INIT_LIST_HEAD(&elmo_host_data.workers);

  elmo_host_data.dmaDescriptors = kmalloc(PAGE_SIZE, GFP_KERNEL);

  printk(KERN_INFO "elmo: allocated DMA buffer at 0x%p, %ld bytes / %d = %ld descriptors\n",
	 elmo_host_data.dmaDescriptors, PAGE_SIZE,
	 sizeof (struct BroadcomDMADescriptor),
	 PAGE_SIZE / sizeof (struct BroadcomDMADescriptor));

  elmo_host_data.llReadBuffer = kmalloc(ELMO_READBUFFER_SIZE, GFP_KERNEL);
  elmo_host_data.llReadBufferHead = elmo_host_data.llReadBufferTail = 0;

  elmo_host_data.llWriteBuffer = kmalloc(ELMO_READBUFFER_SIZE, GFP_KERNEL);
  elmo_host_data.llWriteBufferHead = elmo_host_data.llWriteBufferTail = 0;

  sema_init(&elmo_host_data.writeSem, ELMO_MAX_PENDING_WRITERS);

  elmo_host_data.bulkFifoSize = ELMO_BULK_FIFO_SIZE;

  /*
    Create and kick off a couple of workers to service the
    read and write queues
  */

  printk(KERN_INFO "elmo: initializing workers\n");

  worker1 = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					   GFP_KERNEL);
  worker2 = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					   GFP_KERNEL);
  worker3 = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					   GFP_KERNEL);
  if (!worker1 || !worker2 || !worker3) {
    printk(KERN_ERR "elmo: could not create FIFO service worker\n");
    return;
  }
  memset(worker1, 0, sizeof *worker1);
  memset(worker2, 0, sizeof *worker2);
  memset(worker3, 0, sizeof *worker3);
  worker1->sn = ++elmo_host_data.workerSN;
  worker2->sn = ++elmo_host_data.workerSN;
  worker3->sn = ++elmo_host_data.workerSN;
  elmo_worker_next_state (worker1, ELMO_WORKER_SERVICE_READ_FIFO);
  elmo_worker_next_state (worker2, ELMO_WORKER_SERVICE_WRITE_FIFO);
  elmo_worker_next_state (worker3, ELMO_WORKER_AWAIT_RESET);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker1, &elmo_host_data.workers);
  list_add_tail((struct list_head *) worker2, &elmo_host_data.workers);
  list_add_tail((struct list_head *) worker3, &elmo_host_data.workers);
  restore_flags(flags);
}

/*
  elmo_probe_bridge - see if the bridge PLD has been programmed.

  Check the bridge PLD to see if it's functional.  This is done by putting
  the Broadcom localbus into 16-bit 8-wait-state mode (so it won't
  hang if the PLD doesn't respond), writing a couple of scratch locations
  in the bridge and then reading them back.  If the values read back
  correctly, the bridge is assumed to be functional, and further ELMO
  probing and initialization can be initiated.

  Do NOT call this function after the ELMO has been initialized!  Doing
  so will probably yngvi up the bridge.

  Returns 1 if the bridge is alive, 0 if not.

*/

static int elmo_probe_bridge(void)
{

#define BCM_BGE_TEST_MODE __SYSREG(0xF0)

  unsigned short registerValue, revision, revisionInverted;

  printk(KERN_INFO "elmo: probing for bridge\n");

  mdelay(10);

  // Setup the 7317 local bus so that Chip Select 3
  // is a 16-bit device with 8 clocks of waitstate.
  // This allows us to attempt to access the bridge's
  // revision registers, without hanging the system
  // if the bridge is non copos mentus.

  *BCM_EBI_CS3_CTRL = 0x0811FB11U;

  barrier();

  // Read the revision value twice (possible FPGA bug)

  registerValue = ELMO_BRIDGE->revision;
  barrier();
  registerValue = ELMO_BRIDGE->revision;
  barrier();
  revisionInverted = 0xFF & (registerValue >> 8);
  revision = 0xFF & registerValue;
  barrier();

  if ((revision ^ revisionInverted) == 0xFF && registerValue != 0xFF00) {
    printk(KERN_INFO "elmo: detected bridge, revision %02X, codeword %04X\n",
	   revision, registerValue);

    mdelay(10);

    // Bridge is up and working OK.  Set CS3 to 16 bits with positive
    // acknowledgement required.

    *BCM_EBI_CS3_CTRL = 0x08004291U;
    WriteRegister(BCM_SPI_EBI_REQ_MUX,0,1);

    WriteRegister(BCM_GPIO_SEL1, ReadRegister( BCM_GPIO_SEL1, 1 ) | 0x43, 1 );
    /* 7317A0 has a bug in the receive, with hw flow control so, we use sw 
     * flow control. We could use hw flow control in the xmit case, but the
     * EBI dma engine cannot be disabled, so we cannot switch between 
     * enabling & disabling external DMA on EBI.
     */
    if (IS_7317A0()) {
        printk("elmo: detected 7317 A0\n");
        /* start EBI Tx, half-word, nonincr. sw flow control */
        *BCM_EBI_DMA_CTRL = 0x01000130; 
    }else {
        printk("elmo: detected 7317 A%d\n",(CHIP_ID & 0xFFFFU));
        /* start EBI Tx, external DMA, word, nonincr. */
        *BCM_EBI_DMA_CTRL = 0xC1002030; 
    }

    *BCM_EBI_DMA_TX_START_ADDR_1 = ELMO_IDE_FIFO_ADDR;
    *BCM_EBI_DMA_TX_BUFFER_SIZE_1 = 0;

    *BCM_EBI_DMA_RX_START_ADDR_1 = ELMO_IDE_FIFO_ADDR;
    *BCM_EBI_DMA_RX_BUFFER_SIZE_1 = 0;

    *BCM_DMA_CH1_MAX_BURST = MAX_ISB_BURST/4; /* # words */
    *BCM_DMA_CH2_MAX_BURST = MAX_ISB_BURST/4; /* # words */ 

    *BCM_DMA_BYTE_REVERSE = 0x00000000;
    *BCM_DMA_HALF_REVERSE = 0x00000000;

    *BCM_DMA_CH1_CONFIG = 0x00000004; /* ISB DMA, chaining mode */
    *BCM_DMA_CH2_CONFIG = 0x00000004; /* ISB DMA, chaining mode */

    *BCM_DMA_CH1_INT_MASK = 4;
    *BCM_DMA_CH2_INT_MASK = 4;

    ELMO_BRIDGE->dma_control = 0;
    ELMO_BRIDGE->ide_waterline = 0x4417U;

    barrier();
  } else {
    printk(KERN_INFO "elmo: bridge not detected (not yet initialized?) - read %04X\n",
	   registerValue);
    mdelay(10);
    return -1;
  }

  // Enable interrupts from ELMO and bridge.  Don't enable CS8420 yet

  if (debug) printk(KERN_INFO "elmo: initializing ELMO bridge interrupts\n");

#ifndef SIMULATE_IRQ
  disable_irq(ELMO_IRQ);
#endif

  mdelay(10);
  
  ELMO_BRIDGE->imr = 0;

  if (debug) printk(KERN_INFO "elmo: bridge initialization completed\n");
  mdelay(10);

  // Kick the Elmo task into activity to start servicing the queues
  
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);

  return 0;

}

static void elmo_bridge_is_alive()
{
  struct list_head *this, *next;
  struct elmo_worker *worker;
  elmo_host_data.pldLoaded = 1;
  notifier_installed = 1;
  ELMO_BRIDGE->stall = 0xFF; /* maximum timeout for stall */
  ELMO_BRIDGE->control = ELMO_CONTROL_LL_FIFO_RST | ELMO_CONTROL_IDE_FIFO_RST;
  ELMO_BRIDGE->control = ELMO_CONTROL_EBI_TIMEOUT_EN;
  ELMO_BRIDGE->imr = 
    ELMO_ISR_BUS_TIMEOUT    | ELMO_ISR_DMA_RD_IRQ     |
    ELMO_ISR_DMA_WR_IRQ     | ELMO_ISR_INT_FROM_ESS   | 
    ELMO_ISR_EBI_TO_ESS_OVR | ELMO_ISR_EBI_TO_ESS_UDR |
    ELMO_ISR_ESS_TO_EBI_OVR | ELMO_ISR_ESS_TO_EBI_UDR |
    ELMO_ISR_EBI_TO_IDE_OVR | ELMO_ISR_EBI_TO_IDE_UDR |
    ELMO_ISR_IDE_TO_EBI_OVR | ELMO_ISR_IDE_TO_EBI_UDR;
#ifdef SIMULATE_IRQ
  printk(KERN_INFO "elmo: starting polled simulation of IRQs\n");
  runIRQs = 1;
  queue_task(&elmo_timer_task_struct, &tq_timer);
#else
  enable_irq(ELMO_IRQ);
#endif
  list_for_each_safe(this,next,&elmo_host_data.workers) {
    worker = (struct elmo_worker *) this;
    worker->go = 1;
  }
}

/*
 *	elmo_init_broadcom - set up the Broadcom CS3 chip-select address maps.
 */

static void elmo_init_broadcom (void)
{
  /* setup CS3 address space:
     base address TEI_BASE, LSBs specify size 1MB (7)
     8 bit device
     self terminating (sync = 0)
     8 wait states, 1 hold, 1 setup
  */
  *BCM_EBI_CS3_BASE = TEI_BASE + 7U;
  *BCM_EBI_CS3_CTRL = 0x0811FB01U;
}


int __init init_module(void)
{
  int result;

  printk(KERN_INFO "elmo: version %s loaded\n", version);

#if 0
  /* With change 165066 ,the broadcom driver assigns memory client priorities 
   * for different * platforms in bcm_linux.c so we dont do anything here
   */
  *BCM_MTN_CLIENT_INDEX    = BCM_ISB_MMC_CLIENT_ID;
  barrier();
  *BCM_MTN_INDEX_DATA_WR   = 0x652D;
  barrier();
#endif

#ifndef SIMULATE_IRQ

  printk(KERN_INFO "elmo: trying to request bridge IRQ %d\n", ELMO_IRQ);
  
  if (request_irq(ELMO_IRQ, do_elmo_intr, SA_INTERRUPT, "Elmo", NULL)) {
    printk(KERN_INFO "elmo: failed to acquire IRQ %d\n", ELMO_IRQ);
    return -ENOMEM;
  }

  printk(KERN_INFO "elmo: trying to request ISB DMA IRQ %d\n",
	 BCM_LINUX_DMA_IRQ);
  
  if (request_irq(BCM_LINUX_DMA_IRQ, broadcom_isb_intr, SA_INTERRUPT, "Elmo", NULL)) {
    printk(KERN_INFO "elmo: failed to acquire IRQ!\n");
    result = -EINVAL;
    goto fail_1;
  }
#endif

  elmo_init();

  init_waitqueue_head( &readerWaitQueue );
  init_waitqueue_head( &bulkIoWaitQueue );

  SET_MODULE_OWNER(&elmo_ll_fops);

  result = register_chrdev(ELMO_MAJOR, "elmo_ll", &elmo_ll_fops);
  if (result < 0) {
    printk(KERN_ERR "elmo_ll: can't get major %d\n",ELMO_MAJOR);
    result = -EINVAL;
    goto fail_2;
  }

  elmo_proc_entry = create_proc_entry("elmo", 0, &proc_root);
  if (elmo_proc_entry) {
      elmo_proc_entry->owner = THIS_MODULE;
      elmo_proc_entry->read_proc = elmo_proc_info;
  }else {
      result = -ENODEV;
      printk("cannot register proc entry\n");
      goto fail_1;
  }

  elmo_init_broadcom();

  printk(KERN_INFO "elmo: initializing tasks\n");
  elmo_task_struct.routine = elmo_task;
  elmo_task_struct.data = NULL;
  elmo_timer_task_struct.routine = elmo_timer_task;
  elmo_timer_task_struct.data = NULL;

  if (!elmo_probe_bridge()) {
    elmo_bridge_is_alive();
  }

  return 0;
fail_1:
  printk("fail_1\n");
  free_irq(ELMO_IRQ, NULL);
  return result;
fail_2:
  printk("fail_2\n");
  free_irq(BCM_LINUX_DMA_IRQ, NULL);
  return result;
}

__exit void cleanup_module(void)
{
  free_irq(ELMO_IRQ, NULL);
  free_irq(DMA_IRQ, NULL);
  if (0 != unregister_chrdev(ELMO_MAJOR, "elmo_ll")) {
    printk(KERN_ERR "elmo_ll unregister failed\n");
  }
#ifdef SIMULATE_IRQ
  runIRQs = 0;
#endif
  remove_proc_entry("elmo", &proc_root);
}

/* XXXX NOT SURE WHERE TO PUT THESE, THEY WOULD BEST FIT IN THE SCp */
static AlignmentBuffer *alignmentBufferArray;
static unsigned int alignmentBufferCount;
struct scatterlist *aligned_iovec;

#ifdef DEBUG_IOVEC
static void hexdump ( unsigned char * ptr, int len )
{
    int fmt = 0;

    while ( len-- )
    {
        if ( 16 == fmt )
        {
            fmt = 0;

            printk ( "\n" );
        }

        fmt++;
        printk ( "%02x ", *ptr++ );
    }

    if ( 0 != fmt )
    {
        printk ( "\n" );        
    }
}

static void
dump_iovec(struct scatterlist* iovec, unsigned count)
{
    unsigned i;
    for (i = 0; i < count ; i++ ) {
        printk( "iovec[%d].address = %p, length = 0x%x\n", 
                i, 
                iovec[i].address,
                iovec[i].length );
        hexdump( (unsigned char*)iovec[i].address, 16 );
    }
}
#endif

static void
CopyAlignmentDataFromUserBuffers(AlignmentBuffer* alignmentBufferArray,
                                 unsigned count)
{
    unsigned i, j;
    unsigned char* dataPtr;
    for (i = 0; i < count; i++) {
        memcpy(alignmentBufferArray[i].data, 
               alignmentBufferArray[i].headAddr,
               alignmentBufferArray[i].headLen);
        dataPtr = alignmentBufferArray[i].data + 
            alignmentBufferArray[i].headLen;
        for (j = 0; alignmentBufferArray[i].tailAddr[j] != NULL; j++) {
            memcpy(dataPtr,
                   alignmentBufferArray[i].tailAddr[j],
                   alignmentBufferArray[i].tailLen[j]);
            dataPtr += alignmentBufferArray[i].tailLen[j];
        }
    }
}

static void
CopyAlignmentDataBackToUserBuffers(AlignmentBuffer* alignmentBufferArray,
                                   unsigned count)
{
    unsigned i, j;
    unsigned char* dataPtr;
    for (i = 0; i < count; i++) {
        memcpy(alignmentBufferArray[i].headAddr,
               alignmentBufferArray[i].data, 
               alignmentBufferArray[i].headLen);
        dataPtr = alignmentBufferArray[i].data + 
            alignmentBufferArray[i].headLen;
        for (j = 0; alignmentBufferArray[i].tailAddr[j] != NULL; j++) {
            memcpy(alignmentBufferArray[i].tailAddr[j],
                   dataPtr,
                   alignmentBufferArray[i].tailLen[j]);
            dataPtr += alignmentBufferArray[i].tailLen[j];
        }
    }
}

static void
InsertAlignmentBuffer(
    AlignmentBuffer* alignmentBufferArray, unsigned* pAlignCount,
    struct scatterlist* pIovec, unsigned short *pIovecSent, 
    struct scatterlist* pThisIovec, unsigned short thisCount,
    unsigned nextIovLen, unsigned *pNextOffset)
{
    /* head data */
    alignmentBufferArray[*pAlignCount].headLen = nextIovLen;
    alignmentBufferArray[*pAlignCount].headAddr =
        (unsigned char*)pIovec[*pIovecSent].address + *pNextOffset;
    
    *pNextOffset = SIZE_ALIGNMENT - nextIovLen;
    (*pIovecSent) ++;

    /* tail data may not be complete in the next buffer, hence the
       loop */
    int j = 0;
    for (j = 0; *pNextOffset != 0; j++, (*pIovecSent) ++) {

        alignmentBufferArray[*pAlignCount].tailLen[j] = *pNextOffset;
        alignmentBufferArray[*pAlignCount].tailAddr[j] =
            (unsigned char*)pIovec[*pIovecSent].address;

        if ( *pNextOffset >= pIovec[*pIovecSent].length ) {
            alignmentBufferArray[*pAlignCount].tailLen[j] = 
                pIovec[*pIovecSent].length;
            *pNextOffset -= pIovec[*pIovecSent].length;
        } else {
            /* we don't advance pIovecSent, only j, since we filled in
              a alignment buffer, but didn't exhaust the current
              iovec. */
            j++;
            break;
        }
    }

    assert ( j < SIZE_ALIGNMENT / 4 );

    alignmentBufferArray[*pAlignCount].tailAddr[j] = NULL;

    pThisIovec[thisCount].address = 
        alignmentBufferArray[*pAlignCount].data;
    pThisIovec[thisCount].length = SIZE_ALIGNMENT;
                
    (*pAlignCount) ++;
}

static void
TakeFromNextIovec(
    struct scatterlist* pIovec, unsigned short *pIovecSent, 
    struct scatterlist* pThisIovec, unsigned thisCount,
    unsigned nextIovLen, unsigned *pNextOffset )
{
    pThisIovec[thisCount].address = 
        (char*)pIovec[*pIovecSent].address + *pNextOffset;

    nextIovLen = nextIovLen / 
        SIZE_ALIGNMENT * SIZE_ALIGNMENT;
    pThisIovec[thisCount].length = nextIovLen;
    
    *pNextOffset += nextIovLen;

    assert ( *pNextOffset <= pIovec[*pIovecSent].length );

    if ( *pNextOffset == pIovec[*pIovecSent].length ) {
        (*pIovecSent) ++;
        (*pNextOffset) = 0;
    }
}

static void
RealignIovec(struct scatterlist* iovec, unsigned count, 
             /* returns */ struct scatterlist** aligned_iovec, 
             unsigned *aligned_count)
{
    // this array does not have to be physically contiguous, hence GFP_KERNEL
    *aligned_iovec = (struct scatterlist*)
        kmalloc(sizeof(struct scatterlist)*count*2, GFP_KERNEL);
    *aligned_count = 0;

    assert( aligned_iovec != NULL );

    /* Allocate the alignment buffers. The room in the iovec is
       already allocated by doubling the number of original entries
       in the iovec, above.

       Alignment buffers must be physically contiguous, hence GFP_ATOMIC
    */
    alignmentBufferArray = (AlignmentBuffer*)
        kmalloc(sizeof(AlignmentBuffer)*count, GFP_KERNEL);
    alignmentBufferCount = 0;    // start with the first alignment buffer

    assert( alignmentBufferArray != NULL );

    /* the next iovec entry to work on */
    unsigned short nextSegment = 0;

    /* the offset within the next iovec entry to start with */
    unsigned nextOffset = 0;
    
    while ( nextSegment < count) {

        /* See if we need to insert an alignment buffer here. We
           do need it only if the remaining buffer size in the
           current buffer is less than SIZE_ALIGNMENT!
        */
        unsigned nextIovLen = iovec[nextSegment].length - nextOffset;

        if ( nextIovLen < SIZE_ALIGNMENT ) {
            /* insert an alignment buffer */
            
            InsertAlignmentBuffer( alignmentBufferArray, &alignmentBufferCount,
                                   iovec, &nextSegment, 
                                   *aligned_iovec, *aligned_count,
                                   nextIovLen, &nextOffset );
            
        } else {

            TakeFromNextIovec( iovec, &nextSegment, 
                               *aligned_iovec, *aligned_count,
                               nextIovLen, &nextOffset );

        }

        (*aligned_count) ++;
    } /* end of while */

}

static void initialize_SCp(struct elmo_request * request)
{
    if (request->sg_count) {
        unsigned aligned_count = 0;
        RealignIovec(request->sg_list, request->sg_count, 
                     &aligned_iovec,&aligned_count);
#ifdef DEBUG_IOVEC
        printk("elmo : original count: %d, aligned count: %d\n",request->sg_count, aligned_count);
        dump_iovec(aligned_iovec, aligned_count);

        printk("physical iovec\n");
        dump_iovec(request->sg_list, request->sg_count);
#endif /* DEBUG_IOVEC */
        if (ELMO_BULK_IO_TO_ESS == request->direction) {
            CopyAlignmentDataFromUserBuffers(alignmentBufferArray,
                                             alignmentBufferCount);
        }
        request->SCp.buffer = aligned_iovec;
        request->SCp.buffers_residual = aligned_count - 1;
        request->SCp.ptr = (char *) request->SCp.buffer->address;
        request->SCp.this_residual = request->SCp.buffer->length;
    } else {
        request->SCp.buffer = NULL;
        request->SCp.buffers_residual = 0;
        request->SCp.ptr = (char *) request->sg_list->address;
        request->SCp.this_residual = request->sg_list->length;
        aligned_iovec = NULL;
        alignmentBufferArray = NULL;
    }

    if (alignmentBufferCount > 0 ) {
        if (debug > 1) printk("elmo: alignmentBufferCount = %d\n", alignmentBufferCount);
    }

    if ( ! elmo_check_host_dma_rules( request ) ) {

      printk( "elmo driver: ERROR!!! Misaligned bulk data path pointers!\n" );

    }
}

/*
  Check the buffer(s) pointed to by a command, and see if all of them
  are compatible with the alignment and size rules for a Broadcom
  DMA transfer.  Every buffer must be 4-byte aligned, and a multiple of
  MAX_ISB_BURST bytes.
*/

static int elmo_check_host_dma_rules (struct elmo_request * request)
{
  int i;
  if (((long int) request->SCp.ptr) % 4 != 0 ||
      request->SCp.this_residual % MAX_ISB_BURST != 0) {
    return 0;
  }
  for (i = 1; i <= request->SCp.buffers_residual; i++) {
    if (((long int) request->SCp.buffer[i].address) % 4 != 0 ||
	request->SCp.buffer[i].length % MAX_ISB_BURST != 0) {
      return 0;
    }
  }
  return 1;
}

/*
  See if there is any more data to be transferred via DMA -
  return 0 if done, or residual size of next buffer if not done
*/

static int elmo_check_more_host_dma (struct elmo_request * request)
{
  while (request->SCp.this_residual == 0 &&
	 request->SCp.buffers_residual > 0) {
    ++request->SCp.buffer;
    --request->SCp.buffers_residual;
    request->SCp.this_residual = request->SCp.buffer->length;
    request->SCp.ptr = request->SCp.buffer->address;
  }
  return request->SCp.this_residual;
}

/*
  Set up the Broadcom ROMbus DMA descriptor array.  Each descriptor
  is limited to [1,1023] words of word-aligned data transfer.
*/

#define ONEDMABURST ( 4096 - MAX_ISB_BURST )

static unsigned long elmoSetupHostDMA (struct elmo_request * request,
				       unsigned long burstSize,
				       int direction,
				       unsigned int *descriptorCount)
{
  const int numDescriptors = PAGE_SIZE / sizeof (struct BroadcomDMADescriptor);
  int currentDescriptor;
  int waterline;
  int thisChunk, transferWordCount;
  unsigned long ioAddress;
  unsigned long consumed;
  currentDescriptor = 0;
  waterline = burstSize;
  memset(elmo_host_data.dmaDescriptors, 0, PAGE_SIZE);
  if (debug) printk(KERN_INFO "elmo: setting up Broadcom descriptors\n");
  ioAddress = ((unsigned long) ELMO_IDE_FIFO) & 0x3FFFFFFC;
  consumed = 0;
  while (waterline > 0 &&
	 currentDescriptor < numDescriptors &&
	 (thisChunk = elmo_check_more_host_dma(request)) > 0) {
    if (debug > 1) printk(KERN_INFO "elmo: want a chunk of %d bytes\n", thisChunk);
    if (thisChunk > waterline) {
      thisChunk = waterline;
      if (debug > 1) printk(KERN_INFO "elmo: clipped chunk to waterline: %d bytes\n", thisChunk);
    }
    if (thisChunk > ONEDMABURST) { /* XXX fix this - dplatt */
      thisChunk = ONEDMABURST;
      if (debug > 1) printk(KERN_INFO "elmo: clipped chunk to %d bytes\n", ONEDMABURST);
    }
    transferWordCount = thisChunk / 4;
    consumed += thisChunk;
    if (debug > 1) printk(KERN_INFO "elmo: want to transfer %d words from 0x%08lX\n",
			  transferWordCount, (unsigned long) request->SCp.ptr);
    elmo_host_data.dmaDescriptors[currentDescriptor].transferByteCount = thisChunk;
    elmo_host_data.dmaDescriptors[currentDescriptor].memoryAddr =
      (0x03FFFFFC & (unsigned int) request->SCp.ptr);
    elmo_host_data.dmaDescriptors[currentDescriptor].f = 
                                        (currentDescriptor == 0);
    elmo_host_data.dmaDescriptors[currentDescriptor].ob = 1;
    if (debug > 1) printk(KERN_INFO "elmo: Broadcom DMA descriptor is %08lX %08lX\n",
			  *(unsigned long *) (&elmo_host_data.dmaDescriptors[currentDescriptor]),
			  *(1 + (unsigned long *) (&elmo_host_data.dmaDescriptors[currentDescriptor])));
    dma_cache_wback_inv((unsigned long) phys_to_virt((unsigned long) request->SCp.ptr), thisChunk);
    request->SCp.ptr           += thisChunk;
    request->SCp.this_residual -= thisChunk;
    waterline                  -= thisChunk;
    currentDescriptor ++;
  }
  if (currentDescriptor == 0) {
    printk(KERN_ERR "elmo: AIEE!  Null ROMbus DMA!\n");
    panic("elmo: Null ROMbus DMA");
  }
  elmo_host_data.dmaDescriptors[currentDescriptor - 1].l = 1;
  if (debug) printk(KERN_INFO "elmo: broadcom descriptors from %p to %p, consume %ld bytes\n",
		    &elmo_host_data.dmaDescriptors[0],
		    &elmo_host_data.dmaDescriptors[currentDescriptor - 1],
		    consumed);
  if (consumed > burstSize) {
      printk(KERN_ERR "elmo: AIEE! attempting to consume more than necessary (%d > %d)\n", consumed, burstSize);
  }
  dma_cache_wback_inv((unsigned long) phys_to_virt((unsigned long) elmo_host_data.dmaDescriptors), PAGE_SIZE);
  /*
    Flush the whole dcache.  Unfortunately, this seems to be necessary
    to ensure that SCSI Generic direct I/O works correctly.  The
    dma_cache_wback_inv() calls above *ought* to be sufficient, I
    think, and they work fine for non-direct I/O, but are not reliable
    for direct I/O.  I suspect that any "dirty" cache lines in
    user-virtual don't get flushed out correctly if direct I/O is
    used, because of the fact that the memory lines are multiply
    mapped (once for the user, and once for the kernel kiobuf).

    Ought to figure out a better way to fix this.  */
  flush_dcache();
  if (debug > 1) printk(KERN_INFO "elmo: set up %d Broadcom descriptors\n", currentDescriptor);
  *descriptorCount = currentDescriptor;
  return consumed;
}

static void elmo_bug(struct elmo_worker *worker, 
		     const char *reason)
{
  printk(KERN_ERR "elmo: I die.\n");
  printk(KERN_ERR "elmo:  %s.\n", reason);
  printk(KERN_ERR "elmo:   %s.\n", bug_line1[jiffies & 0x3]);
  printk(KERN_ERR "elmo:    %s.\n", bug_line2[(jiffies >> 2) & 0x3]);
  printk(KERN_ERR "elmo:     All die.  O the embarrassment!\n");
  if (reason) {
      printk(KERN_ERR "elmo: reason %s\n", reason);
  }
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
}

static void elmo_reset_ATA() {
  unsigned long flags;
  save_and_cli(flags);
  ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
  ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
  barrier();
  ELMO_BRIDGE->control |=  ELMO_CONTROL_DSK_RST_ACK;
  ELMO_BRIDGE->control &= ~ELMO_CONTROL_DSK_RST_ACK;
  barrier();
  elmo_host_data.wasReset = 0;
  printk(KERN_ERR "elmo: ATA FIFO cleared, ATA reset acknowledged\n");
  restore_flags(flags);
}

static struct elmo_worker *elmo_bulk_io_init (struct elmo_request * request)
{
  struct elmo_worker *worker;
  unsigned long flags;
  if (debug) printk(KERN_INFO "elmo: initializing bulk io for xfer %d\n", request->xferId);
  (void) worker;
  (void) flags;
  if (!elmo_host_data.pldLoaded) {
      printk(KERN_ERR "attempting to start builk io on an uninitialized bridge\n");
      return NULL;
  }
  /*
    Create a Worker for this request and queue it up...
  */
  if (debug > 1) printk(KERN_INFO "elmo: allocating worker\n");
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
				       GFP_KERNEL);
  if (debug > 1) printk(KERN_INFO "elmo: got worker at %p\n", worker);
  if (!worker) {
    return NULL;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++elmo_host_data.workerSN;
  worker->request = request;
  if (debug) printk(KERN_INFO "elmo: worker is #%d\n", worker->sn);
  elmo_worker_next_state(worker, ELMO_WORKER_SETUP);
  save_and_cli(flags);
  initialize_SCp(request);
  list_add_tail((struct list_head *) worker, &elmo_host_data.workers);
  /*
    Take it out of the Connected slot and mark us non-busy
  */
  restore_flags(flags);
  return worker;
}


static void elmo_release_chip_resources (unsigned int resourceMask)
{
  struct list_head *this;
  struct elmo_worker *worker;
  unsigned long flags;
  unsigned int releaseMask;
  unsigned int intersectMask;
  releaseMask = resourceMask;
  elmo_host_data.resourcesBusy &= ~resourceMask;
  if (debug > 1) printk(KERN_INFO "elmo: releasing resource mask 0x%08X\n",
			resourceMask);
  save_and_cli(flags);
  list_for_each(this,&elmo_host_data.workers) {
    worker = (struct elmo_worker *) this;
    if (worker->elmoResourcesUsed & resourceMask) {
      worker->elmoResourcesUsed &= ~resourceMask;
    }
    intersectMask = worker->elmoResourcesNeeded & releaseMask;
    if (intersectMask) {
      if (debug > 1) printk(KERN_INFO "elmo: worker %d wants 0x%08X\n",
			    worker->sn, intersectMask);
      worker->elmoResourcesNeeded &= ~intersectMask;
      worker->go = 1;
      /*      intersectMask &= ~ELMO_MULTIPLE_GRABBERS; */
      releaseMask &= ~intersectMask;
    }
  }
  restore_flags(flags);
}

/*
  Other exported routines, for other drivers which must manipulate
  the ELMO bridge registers.
*/

int ELMOReadBridge(unsigned long address, unsigned long length, void *data )
{
  unsigned long flags;
  unsigned long *dataWord;

  if (!elmo_host_data.pldLoaded) {
    return -ENODEV;
  }
  dataWord = (unsigned long *) data;
  save_and_cli(flags);
  while (length > 0) {
    *dataWord = *((unsigned long *) (ELMO_BRIDGE_BASE + address));
    length -= sizeof (unsigned long);
    address += sizeof (unsigned long);
    dataWord += sizeof (unsigned long);
  }
  restore_flags(flags);
  return 0;
}

int ELMOWriteBridge(unsigned long address, unsigned long length, void *data )
{
  unsigned long flags;
  unsigned long *dataWord;

  if (!elmo_host_data.pldLoaded) {
    return -ENODEV;
  }
  dataWord = (unsigned long *) data;
  save_and_cli(flags);
  while (length > 0) {
    *((unsigned long *) (ELMO_BRIDGE_BASE + address)) = *dataWord;
    length -= sizeof (unsigned long);
    address += sizeof (unsigned long);
    dataWord += sizeof (unsigned long);
  }
  restore_flags(flags);
  return 0;
}

int ELMOWriteBridgeMasked(unsigned long address, unsigned short mask,
			unsigned short bits)
{
  unsigned long flags;

  if (!elmo_host_data.pldLoaded) {
    return -ENODEV;
  }
  save_and_cli(flags);
  *((unsigned long *) (ELMO_BRIDGE_BASE + address)) =
    (~mask & *((unsigned long *) (ELMO_BRIDGE_BASE + address))) |
    ( mask & bits);
  restore_flags(flags);
  return 0;
}

/*
  State-machine nodes for the bulk-data transfer channel.
*/

static void elmo_worker_setup(struct elmo_worker *worker)
{
  unsigned long int length, i;
  struct elmo_request *request;  
  request = (struct elmo_request *) worker->request;

  length = request->SCp.this_residual;
  if (debug > 1) printk(KERN_INFO "elmo: buffer %p(%ld)\n", request->SCp.ptr, length);
  for (i = 1; i <= request->SCp.buffers_residual; i++) {
    if (debug > 1) printk(KERN_INFO "elmo: %p(%d)\n", request->SCp.buffer[i].address,
		      request->SCp.buffer[i].length);
    length += request->SCp.buffer[i].length;
  }
  if (debug) printk(KERN_INFO "elmo: setup request, command : %s for %ld bytes\n", (request->direction == ELMO_BULK_IO_TO_ESS) ? "WRITE" : "READ",
	 length);
  worker->direction = request->direction;
  worker->transferSize = length;
  worker->transferRemaining = length;
  worker->residual = length;
  elmo_worker_next_state(worker, ELMO_WORKER_GET_BULK_INTERFACE);
}

static void elmo_worker_done(struct elmo_worker *worker)
{
  struct elmo_request *request;  
  if (debug) printk(KERN_INFO "elmo: worker %d is done\n", worker->sn);
  if (worker->dataBuffer) {
    kfree(worker->dataBuffer);
    worker->dataBuffer = NULL;
  }
  if (worker->holdingSemaphore) {
    up(&elmo_host_data.writeSem);
    if (debug > 1) printk(KERN_INFO "elmo: worker %d up sem to %d\n",
			  worker->sn,
			  sem_getcount(&elmo_host_data.writeSem));
  }
  if (worker->signalATA) {
    /*
      XXX This isn't really right - we'd like to generate an ATA-bus
      "done" event by writing to an ATA-bus register, and perhaps even
      pass along error information if there's a problem.  Alas, the
      FPGA doesn't provide us with access to the necessary registers.
    */
    ELMO_BRIDGE->control |= ELMO_CONTROL_INT_ESS;
    barrier();
    ELMO_BRIDGE->control &= ~ELMO_CONTROL_INT_ESS;
  }
  if (elmo_host_data.wasReset) {
    /*
      Previous command might have left the ATA in a funky state.
      If so, clear the FIFOs out and acknowledge the reset, and
      continue onwards.
    */
    if (debug > 1) printk(KERN_INFO "elmo: worker %d clearing prior reset\n",
			  worker->sn);
    elmo_reset_ATA();
  }
  elmo_release_chip_resources(worker->elmoResourcesUsed);
  request = (struct elmo_request *) worker->request;
  if (request) {
    if (worker->reportHang) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting a data path failure\n");
      worker->request->dio_request->status |= ELMO_ERROR_HANG;
    } else if (worker->reportError) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting an error\n");
      worker->request->dio_request->status |= ELMO_ERROR_UNKNOWN;
    } else if (worker->aborting) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting a command abort\n");
      if (!worker->request->dio_request->status) {
          worker->request->dio_request->status |= ELMO_ERROR_ABORT;
      }
    } else if (worker->errorATA) {
      if (debug > 1) printk(KERN_INFO "elmo: reporting errorATA\n");
      worker->request->dio_request->status |= ELMO_ERROR_ATA;
    }
    if ( ELMO_BULK_IO_FROM_ESS == worker->request->direction ) {
        CopyAlignmentDataBackToUserBuffers(alignmentBufferArray,
                                           alignmentBufferCount);
    }
    if (NULL != alignmentBufferArray) {
        kfree(alignmentBufferArray);
        alignmentBufferArray = NULL;
    }

    if (NULL != aligned_iovec) {
        kfree(aligned_iovec);
        aligned_iovec = NULL;
    }
    if (NULL != request->dio_request) {
        elmo_host_data.user_request = request->dio_request;
    }
    kfree((struct elmo_request *)worker->request);
    elmo_host_data.bulkIOCompleted = 1;

  }

  worker->request = NULL;
  elmo_autodebug_off();  /* turn off autodebug if it has timed out */
  elmo_worker_disarm_deadman(worker);
  if (elmo_host_data.bulkIOCompleted) {
      wake_up(&bulkIoWaitQueue);
  }
}

static void elmo_worker_get_bulk_interface(struct elmo_worker *worker)
{
  if (worker->aborting) {
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  if (!(elmo_host_data.resourcesBusy & ELMO_RESOURCE_BULK_INTERFACE)) {
    worker->elmoResourcesUsed |= ELMO_RESOURCE_BULK_INTERFACE;
    elmo_host_data.resourcesBusy |= ELMO_RESOURCE_BULK_INTERFACE;
    if (debug > 1) printk(KERN_INFO "elmo: worker %d has bulk I/O\n",
			  worker->sn);
    /*
      Now that we own it, make sure we're notified if the ESS
      tries a reset while we're using it
    */
    worker->elmoInterruptsFired &= ~ ELMO_INT_DSK_RST_REQ;
    worker->elmoInterruptsPending |= ELMO_INT_DSK_RST_REQ;
    if (elmo_host_data.wasReset) {
      /*
	Previous command might have left the ATA in a funky state.
	If so, clear the FIFOs out and acknowledge the reset, and
	continue onwards.
      */
      if (debug > 1) printk(KERN_INFO "elmo: worker %d clearing prior reset\n",
			    worker->sn);
      elmo_reset_ATA();
    }
    elmo_worker_next_state(worker, ELMO_WORKER_GET_ESS_COMMAND);
  } else {
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_BULK_INTERFACE;
    if (debug > 1) printk(KERN_INFO "elmo: worker %d queued for bulk I/O\n",
			  worker->sn);
  }
}

static void elmo_worker_get_ess_command(struct elmo_worker *worker)
{
  struct elmo_request *request = (struct elmo_request *) worker->request;
  unsigned char commandByte;
  unsigned int sectorCountByte;
  unsigned short lba;
  unsigned long flags;
  if (worker->aborting) {
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  /*
    If the ESS reset the bus while we were in charge of it, we've
    got bad problems.  Reset the bus and then bail out.
  */
  if (elmo_host_data.wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
     printk(KERN_ERR "elmo: worker %d hit by ATA reset while awaiting command\n", worker->sn);
    elmo_reset_ATA();
    worker->aborting = 1;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  /*
    If the ESS got its command to us before we were really ready for
    it, the disk-command-received interrupt will be trying to fire.
    Save a bit of time by acknowledging and clearing it here, and 
    pretending that it had fired.
  */
  save_and_cli(flags);
  if (!elmo_host_data.haveCommand) {
    /*
      No command yet, wait for it
    */
    if (debug > 1) printk(KERN_INFO "elmo: worker %d queued for ATA command\n",
			  worker->sn);
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_ATA_COMMAND;
    restore_flags(flags);
    PLOG(PLOG_ELMO_BULK_WAIT_FOR_ATA_CMD,0,0);
    return;
  }
  /*
    Command has arrived.  Capture the registers.
  */
  elmo_host_data.haveCommand = 0;
  worker->ataCommandReg = elmo_host_data.ataCommandReg;
  worker->ataCylinderReg = elmo_host_data.ataCylinderReg;
  worker->signalATA = 1;
  restore_flags(flags);
  if (debug > 1) printk(KERN_INFO "elmo: worker %d has ATA command\n",
			worker->sn);
  commandByte = worker->ataCommandReg & 0xFF;
  sectorCountByte = (worker->ataCommandReg >> 8) & 0xFF;
  if (sectorCountByte == 0 && ESS_ATA_TRANSFER_ABORT != commandByte) {
    sectorCountByte = 256;
  }
  PLOG(PLOG_ELMO_BULK_GOT_ATA_CMD,sizeof(commandByte),&commandByte);
  if (debug > 1) printk(KERN_INFO "elmo: worker %d has ATA command %02X for %d sectors\n",
			worker->sn, commandByte, sectorCountByte);
  switch (commandByte) {
  case WIN_READ:
  case WIN_READDMA:
  case WIN_MULTREAD:
    if (request->direction != ELMO_BULK_IO_TO_ESS) {
      elmo_bug(worker, "The ESS asks to read, but we are not writing");
      worker->errorATA = 1;
      return;
    }
    break;
  case WIN_WRITE:
  case WIN_WRITEDMA:
  case WIN_MULTWRITE:
    if (request->direction != ELMO_BULK_IO_FROM_ESS) {
      elmo_bug(worker, "The ESS asks to write, but we are not reading");
      worker->errorATA = 1;
      return;
    }
    break;
  case ESS_ATA_TRANSFER_ABORT:
    {
      struct elmo_request *request = (struct elmo_request *) worker->request;
      if (debug > 1) printk(KERN_INFO "elmo: received vendor specific abort from ESS\n");
      if (request) {
        /*
          Pass back a recognizable error to the caller
        */
        request->dio_request->status |= ELMO_ERROR_ATA_ABORT;
        request->dio_request->client_data = sectorCountByte;
      }
      if (debug > 1) printk(KERN_INFO "elmo: ESS says abort, so aborting bulk data transfer\n");
      worker->aborting = 1;
      elmo_worker_next_state(worker, ELMO_WORKER_DONE);
#if 0
      ELMO_BRIDGE->control |=  ELMO_CONTROL_DSK_CMD_ACK;
      ELMO_BRIDGE->control &= ~ELMO_CONTROL_DSK_CMD_ACK;
#endif
      return;
    }
  default:
    elmo_bug(worker, "The ESS sends a command we do not understand");
    worker->errorATA = 1;
    return;
  }
  lba = request->xferId;
  if (lba != worker->ataCylinderReg) {
    printk( KERN_ERR
	    "elmo: TiVo expecting ATA transfer ID %d, ESS asking for %d\n",
	    lba, worker->ataCylinderReg );
    elmo_bug(worker, "The command sequence numbers do not match");
    worker->errorATA = 1;
    return;
  } else if (sectorCountByte * 512 > worker->transferRemaining) {
    printk( KERN_ERR
	    "elmo: TiVo ready to transfer %d bytes, ESS wants %d bytes\n", 
	    worker->transferRemaining, sectorCountByte * 512 );
    elmo_bug(worker, "The ESS wants to transfer too much data");
    worker->errorATA = 1;
    return;
  } 
  worker->transferBurst = sectorCountByte * 512;
  worker->transferRemaining -= worker->transferBurst;
  /*
      We've got the command and are ready to acknowledge it so the
      ESS can start the data transfer.  If they're doing a write,
      we can't start our DMA immediately - it would stall on a PY - so
      we queue up for the data-in-FIFO interrupt and go to sleep.
      If they're doing a read, we can go set up our DMA now.
    */
  if (commandByte == WIN_READ || 
      commandByte == WIN_READDMA || 
      commandByte == WIN_MULTREAD) {
    if (debug > 1) printk(KERN_INFO "elmo: can go push data to FIFO\n");
#ifdef TEMPORARY_SLOW_DMA_HACK
    elmo_queue_worker_for_timer(worker, 10, ELMO_WORKER_SETUP_DMA);
#else
    elmo_worker_next_state(worker, ELMO_WORKER_SETUP_DMA);
#endif
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: must wait for data in FIFO\n");
    elmo_queue_worker_for_interrupt (worker, ELMO_INT_IDE_TO_EBI_NE, 
				     ELMO_WORKER_AWAIT_FIFO);
    ELMO_BRIDGE->fmr |= ELMO_FIFO_IDE_TO_EBI_NE;
  }
}

static void elmo_worker_await_fifo(struct elmo_worker *worker)
{
  if (elmo_host_data.wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
     printk(KERN_ERR "elmo: worker %d hit by ATA reset while awaiting FIFO\n",
	    worker->sn);
    elmo_reset_ATA();
    worker->aborting = 1;
  }
  if (worker->aborting) {
    worker->errorATA = 1;
    ELMO_BRIDGE->fmr &= ~ELMO_FIFO_IDE_TO_EBI_NE;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  if (worker->elmoInterruptsFired & ELMO_INT_IDE_TO_EBI_NE) {
    ELMO_BRIDGE->fmr &= ~ELMO_FIFO_IDE_TO_EBI_NE;
#ifdef TEMPORARY_SLOW_DMA_HACK
      elmo_queue_worker_for_timer(worker, 10, ELMO_WORKER_SETUP_DMA);
#else
      elmo_worker_next_state(worker, ELMO_WORKER_SETUP_DMA);
#endif
  } else {
    printk(KERN_INFO "elmo: worker awaiting IDE FIFO woke for no reason\n");
  }
}

static void elmo_worker_setup_dma(struct elmo_worker *worker)
{
  struct elmo_request *request = (struct elmo_request *) worker->request;
  unsigned long consumed;
  unsigned long flags;
  unsigned short watermarkBit;
  unsigned long watermarkInterrupt;
  unsigned int direction;
  unsigned int transferBurst;
  unsigned int descriptorCount;
  unsigned int transferCount = 0;
  unsigned int myDMA = 0;

  if (worker->direction == ELMO_BULK_IO_FROM_ESS) {
    direction = 0;
    watermarkBit = ELMO_FIFO_IDE_TO_EBI_HWM;
    watermarkInterrupt = ELMO_INT_IDE_TO_EBI_HWM;
  } else {
    direction = 1;
    watermarkBit = ELMO_FIFO_EBI_TO_IDE_LWM;
    watermarkInterrupt = ELMO_INT_EBI_TO_IDE_LWM;
  }
  save_and_cli(flags);
  ELMO_BRIDGE->fmr &= ~watermarkBit;
  worker->elmoInterruptsFired &= ~ watermarkInterrupt;
  worker->elmoInterruptsPending &= ~ watermarkInterrupt;
  restore_flags(flags);
  if (elmo_host_data.wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
     printk(KERN_ERR "elmo: worker %d hit by ATA reset while prepping DMA\n",
	    worker->sn);
    elmo_reset_ATA();
    worker->aborting = 1;
  }
  if (worker->aborting) {
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    return;
  }
  if (worker->transferBurst == 0) {
    if (worker->residual == 0) {
      if (debug) printk(KERN_INFO "elmo: all data has been transferred\n");
      elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    } else {
      if (debug) printk(KERN_INFO "elmo: more data to transfer\n");
      elmo_worker_next_state(worker, ELMO_WORKER_GET_ESS_COMMAND);
    }
    return;
  }
  PLOG(PLOG_ELMO_BULK_DMA_START,0,0);
  save_and_cli(flags);
  if (!(ELMO_BRIDGE->fsr & watermarkBit)) {
    elmo_queue_worker_for_interrupt (worker, watermarkInterrupt,
				     ELMO_WORKER_SETUP_DMA);
    ELMO_BRIDGE->fmr |= watermarkBit;
    if (debug) printk(KERN_INFO "elmo: awaiting FIFO watermark\n");
    restore_flags(flags);
    return;
  } 
  restore_flags(flags);
  transferBurst = worker->transferBurst;
  if (IS_7317A0()) {
      /* 7317A0 has a bug in the receive, with hw flow control so, we use sw flow
       * control. We could use hw flow control in the xmit case, but the EBI dma
       * engine cannot be disabled, so we cannot switch between enabling &
       * disabling external DMA on EBI.
       */
      if (elmo_host_data.bulkFifoSize != 0 &&
          transferBurst > elmo_host_data.bulkFifoSize / 2) {
        transferBurst = elmo_host_data.bulkFifoSize / 2;
      }
  }

  consumed = elmoSetupHostDMA (request, transferBurst, direction,
			       &descriptorCount);
  worker->consuming = consumed;
  worker->transferBurst -= consumed;
  transferCount = consumed/(MAX_ISB_BURST); 
  /*
    Set up the Broadcom DMA registers and start the engine.
  */
  save_and_cli(flags);
  if (direction) {
    if (debug) printk(KERN_INFO "elmo: starting Broadcom write-DMA burst\n");

    *BCM_DMA_CH2_BUFF_ADDR = 
      (((unsigned long) elmo_host_data.dmaDescriptors) & 0x03FFFFFF);
    *BCM_DMA_CH2_BCNTR_LEN = descriptorCount;

    if (IS_7317A0()) {
        *BCM_EBI_DMA_TX_BUFFER_SIZE_1 = consumed;
    }

    if (0 != swaponwrite) {
        ELMO_BRIDGE->dma_control = 0x8000U; 
    }else {
        ELMO_BRIDGE->dma_control = 0;
    }
    barrier();


    elmo_broadcom_isb_enabled = 1;

    worker->dmaRunning = ELMO_DMA_WRITING;

    if (!IS_7317A0()) {
        myDMA = ELMO_INT_DMA_WR_IRQ;
        elmo_asic_write_dma_enabled = 1;
    }else {
        myDMA = 0;
        elmo_asic_write_dma_enabled = 0;
    }


    *BCM_EBI_DMA_CTRL |= 0x00000032U; /* start EBI Tx, word, nonincr. */
    barrier();
    /* enable write dma flow control */
    if (!IS_7317A0()) {
        ELMO_BRIDGE->dma_control |= (0x2000U|(transferCount&0xFFFU)); 
       barrier();
    }
    *BCM_DMA_CH2_CONFIG |= 0x00000001U; /* start ISB DMA, chaining mode */
    barrier();
  } else {

    if (debug) printk(KERN_INFO "elmo: starting Broadcom read-DMA burst\n");
    *BCM_DMA_CH1_BUFF_ADDR =
      (((unsigned long) elmo_host_data.dmaDescriptors) & 0x03FFFFFF);
    *BCM_DMA_CH1_BCNTR_LEN = descriptorCount;

    if (IS_7317A0()) {
        *BCM_EBI_DMA_RX_BUFFER_SIZE_1 = consumed;
    }
    if (0 != swaponread) {
        ELMO_BRIDGE->dma_control = 0x4000U; 
    }else {
        ELMO_BRIDGE->dma_control = 0;
    }

    barrier();

    *BCM_DMA_CH1_CONFIG |= 0x00000001U; /* start ISB DMA, chaining mode */
    *BCM_EBI_DMA_CTRL   |= 0x00000031U; /* start EBI Rx, word, nonincr. */
    barrier();

    elmo_broadcom_isb_enabled = 1;

    worker->dmaRunning = ELMO_DMA_READING;

    if (!IS_7317A0()) {
        myDMA = ELMO_INT_DMA_RD_IRQ;
        elmo_asic_read_dma_enabled = 1;
    }else {
        myDMA = 0;
        elmo_asic_read_dma_enabled = 0;
    }

    /* enable read dma flow control */
    if (!IS_7317A0()) {
        ELMO_BRIDGE->dma_control |=(0x1000U|(transferCount&0xFFFU)); 
       barrier();
    }
  }

  elmo_queue_worker_for_interrupt (worker, myDMA |
                                   ELMO_INT_ISB_DMA|
				   ELMO_INT_BUS_TIMEOUT |
				   ELMO_INT_EBI_TO_ESS_OVR |
				   ELMO_INT_ESS_TO_EBI_UDR,
				   ELMO_WORKER_FINISH_DMA);
  enable_irq(BCM_LINUX_DMA_IRQ);
  restore_flags(flags);

#if 0
#define FAKE_ATA_RESET_FROM_ESS
#ifdef FAKE_ATA_RESET_FROM_ESS
  if (!direction) {
    ELMO_BRIDGE->control |=  ELMO_CONTROL_ATA_INT_RST;
    ELMO_BRIDGE->control &= ~ELMO_CONTROL_ATA_INT_RST;
    barrier();
  }
#endif /* FAKE_ATA_RESET_FROM_ESS */
#endif /* 0 */
}

static void elmo_stop_isb_dma( volatile u32* isb_config_reg)
{
    int count = 10000;

    if (!elmo_broadcom_isb_enabled) {
        return;
    }

    /* stop ISB DMA */
    do {
        barrier();

        *isb_config_reg |= 0x00000002U; /* stall ISB DMA */

        barrier();
        udelay(1000);
    }while((!(*isb_config_reg & 0x00000002U)) && count--);

    if (!count) {
        printk("elmo : unable to stall isb dma\n");
        panic("elmo : unable to stall isb dma\n");
    }

    barrier();

    *isb_config_reg = 0x00000040U;
    barrier();

    count = 10000;

    do {
        barrier();
        *isb_config_reg &= 0xFFFFFFFEU; /* stop ISB DMA */
        barrier();
        udelay(1000);
    }while((*isb_config_reg & 0x00000001U) && count--);

    if (!count) {
        printk("elmo : unable to stop isb dma\n");
        panic("elmo : unable to stop isb dma\n");
    }

    *isb_config_reg = 0x00000004U;
    elmo_broadcom_isb_enabled = 0;
}

static void elmo_worker_finish_dma(struct elmo_worker *worker)
{
  unsigned long flags;
  unsigned int resetATA = 0;
  unsigned int finished = 0;
  PLOG(PLOG_ELMO_BULK_DMA_END,0,0);
  save_and_cli(flags);
  if (elmo_host_data.wasReset ||
      (worker->elmoInterruptsFired & ELMO_INT_DSK_RST_REQ)) {
    printk(KERN_ERR "elmo: worker %d hit by ATA reset while doing DMA\n",
	   worker->sn);
    resetATA = 1;
    worker->aborting = 1;
  }
  if (worker->elmoInterruptsFired & ELMO_INT_BUS_TIMEOUT) {
    printk(KERN_ERR "elmo: Oh bother.  Worker %d stalled the Broadcom DMA\n",
		      worker->sn);
    worker->aborting = 1;
    worker->reportHang = 1;
  }
  if (worker->elmoInterruptsFired & (ELMO_INT_EBI_TO_ESS_OVR |
				     ELMO_INT_ESS_TO_EBI_UDR)) {
    printk(KERN_ERR "elmo: Oh fudge.  Worker %d thrashed the FIFO\n",
	   worker->sn);
    worker->aborting = 1;
  }
  if (worker->aborting) {
    switch (worker->dmaRunning) {
    case ELMO_DMA_IDLE:
      break;
    case ELMO_DMA_READING:
        if (elmo_broadcom_isb_enabled) {
            printk(KERN_INFO "elmo: aborting DMA read\n");

            if (elmo_asic_read_dma_enabled) {
                /* set the FPGA DMA transfer count to 0 but let it run.
                 * This will ensure that the current burst is transferred and
                 * we will not have data stuck in the any of the FIFOs. Stop
                 * the FPGA DMA after ISB has been stopped
                 */
                ELMO_BRIDGE->dma_control &=~(0xFFFU);
                barrier();
            }

            elmo_stop_isb_dma(BCM_DMA_CH1_CONFIG);

            /* stop EBI RX DMA. 
             * This will only be effective on a 7317-A0 system
             */
            *BCM_EBI_DMA_CTRL   &= ~0x00000001U; 
            barrier();
            *BCM_EBI_DMA_CTRL   |= 0x00000030U; /* clear EBI interrupt flags */

            if (elmo_asic_read_dma_enabled) {
                /* stop the FPGA DMA */
                ELMO_BRIDGE->dma_control &=~(0x1000U);
                barrier();
                elmo_asic_read_dma_enabled = 0;
            }

            ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
            ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
            barrier();

      }
      worker->dmaRunning = ELMO_DMA_IDLE;
      break;
    case ELMO_DMA_WRITING:
      if ( elmo_asic_write_dma_enabled || 
           ( IS_7317A0() && elmo_broadcom_isb_enabled )) {
	printk(KERN_INFO "elmo: aborting DMA write\n");

        /* stop ISB DMA */
        elmo_stop_isb_dma(BCM_DMA_CH2_CONFIG);

        /* reset the FPGA FIFO to make room for any pending ISB DMA to 
         * go through 
         */
        ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
        ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
        barrier();

        /* set the FPGA DMA transfer count to 0 but let it run.
         * This will ensure that the current burst is transferred and
         * we will not have data stuck in the any of the FIFOs. Stop
         * the FPGA DMA after ISB has been stopped
         */
        ELMO_BRIDGE->dma_control &=~(0xFFFU);
        barrier();

        /* stop EBI TX DMA. 
         * This will only be effective on a 7317-A0 system
         */
        *BCM_EBI_DMA_CTRL   &= ~0x00000002U;
        barrier();
        *BCM_EBI_DMA_CTRL   |= 0x00000030U; /* clear EBI interrupt flags */

        /* stop the FPGA DMA */
        ELMO_BRIDGE->dma_control &=~(0x2000U);
        barrier();
        /* reset the FPGA FIFOs again */
        ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
        ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
        barrier();
	elmo_asic_write_dma_enabled = 0;
      }
      worker->dmaRunning = ELMO_DMA_IDLE;
      break;
    }
    if (resetATA) {
      elmo_reset_ATA();
    }
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    restore_flags(flags);
    return;
  }
  switch (worker->dmaRunning) {
  case ELMO_DMA_IDLE:
  default:
    printk(KERN_INFO "elmo: worker %d lost state of Broadcom DMA\n",
	   worker->sn);
    worker->aborting = 1;
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    break;
  case ELMO_DMA_READING:
    if (!elmo_asic_read_dma_enabled && !elmo_broadcom_isb_enabled) {
      finished = 1;
    }
    if (debug) printk(KERN_INFO "elmo: worker %d finds read DMAs %sdone\n",
		      worker->sn, finished ? "" : "not ");
    break;
  case ELMO_DMA_WRITING:
    if (!elmo_asic_write_dma_enabled && !elmo_broadcom_isb_enabled) {
      finished = 1;
    }
    if (debug) printk(KERN_INFO "elmo: worker %d finds write DMAs %sdone\n",
		      worker->sn, finished ? "" : "not ");
      break;
  }
  if (finished) {
    if (debug) printk(KERN_INFO "elmo: worker %d ends Broadcom DMA\n",
		      worker->sn);
    worker->residual -= worker->consuming;
    elmo_worker_next_state(worker, ELMO_WORKER_SETUP_DMA);
  }
  restore_flags(flags);
}

static struct elmo_worker * elmo_abort_bulk_worker()
{
  unsigned long flags;
  struct list_head *this, *next;
  struct elmo_worker *bulk_worker = NULL;

  list_for_each_safe(this,next,&elmo_host_data.workers) {
    struct elmo_worker *tmp_worker = (struct elmo_worker *) this;
    if (tmp_worker->elmoResourcesUsed & ELMO_RESOURCE_BULK_INTERFACE) {
        bulk_worker = tmp_worker;
        break;
    }
  }

  if (NULL == bulk_worker) {
      if (debug > 1) printk(KERN_INFO "elmo: cannot find worker performing bulk IO.\n");
      elmo_release_chip_resources(ELMO_RESOURCE_ATA_COMMAND);
      return NULL;
  }

  save_and_cli(flags);
  elmo_host_data.haveCommand = 0;
  bulk_worker->aborting = 1;
  bulk_worker->go = 1;
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  restore_flags(flags);

  return bulk_worker;
}

static void elmo_worker_await_reset(struct elmo_worker *worker)
{
  unsigned long flags, fired;
  int i;
  save_and_cli(flags);
  fired = worker->elmoInterruptsFired;
  worker->elmoInterruptsFired = 0;
  restore_flags(flags);
  PLOG(PLOG_ELMO_BULK_WAITING_FOR_CMD,0,0);
  if (fired & ELMO_INT_DSK_RST_REQ) {
    if (debug > 1) printk(KERN_INFO "elmo: shmoo sees that ATA reset has arrived\n");
    elmo_host_data.haveCommand = 0; /* previous command voided! */
    if (elmo_host_data.resourcesBusy & ELMO_RESOURCE_BULK_INTERFACE) {
      /*
	Another worker is in charge of the bus.  Make sure that this
	worker sees the reset, and acknowledges it, before releasing
	the bus.
      */
      elmo_host_data.wasReset = 1;
    } else {
      /*
	No one has the bus at the moment.  Acknowledge the reset now
	so that the AT-host can unblock.
      */
      elmo_reset_ATA();
    }
  }
  if (fired & ELMO_INT_DSK_CMD_REQ) {
    if (debug > 1) printk(KERN_INFO "elmo: shmoo sees that ATA command has arrived\n");
    if (elmo_host_data.haveCommand) {
      printk(KERN_ERR "elmo: command arrived with command pending!\n");
    }
    elmo_host_data.haveCommand = 1;
    elmo_host_data.ataCommandReg = ELMO_BRIDGE->command;
    elmo_host_data.ataCylinderReg = ELMO_BRIDGE->cylinder;
    /*
      Acknowledge the command, turn the ESS loose.
    */
    ELMO_BRIDGE->control |=  ELMO_CONTROL_DSK_CMD_ACK;
    barrier();
    ELMO_BRIDGE->control &= ~ELMO_CONTROL_DSK_CMD_ACK;
    PLOG(PLOG_ELMO_BULK_ACKNOWLEDGED_CMD,sizeof(elmo_host_data.ataCommandReg),
            &elmo_host_data.ataCommandReg);
    switch (elmo_host_data.ataCommandReg & 0xFF) {
    case WIN_READ:
    case WIN_READDMA:
    case WIN_WRITE:
    case WIN_WRITEDMA:
    case WIN_MULTWRITE:
    case WIN_MULTREAD:
      elmo_release_chip_resources(ELMO_RESOURCE_ATA_COMMAND);
      break;
    case WIN_IDENTIFY:
      printk(KERN_INFO "elmo: performing dummy IDENTIFY\n");
      save_and_cli(flags);
      for (i = 0; i < 32; i++) {
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x4E75;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x6D62;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x6572;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x206E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x696E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x652E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x2E2E;
	*((volatile short *)(ELMO_IDE_FIFO_ADDR)) = 0x2020;
      }
      elmo_host_data.haveCommand = 0;
      restore_flags(flags);
      break;
  case ESS_ATA_TRANSFER_ABORT:
    {
      struct elmo_worker *bulk_worker = elmo_abort_bulk_worker();
      if (NULL == bulk_worker) {
          break;
      }

      struct elmo_request *request = (struct elmo_request *) bulk_worker->request;

      if (request) {
        /*
          Pass back a recognizable error to the caller
        */
        request->dio_request->status |= ELMO_ERROR_ATA_ABORT;
        unsigned int sectorCountByte = 
            (bulk_worker->ataCommandReg >> 8) & 0xFF;
        request->dio_request->client_data = sectorCountByte;
      }
      if (debug > 1) printk(KERN_INFO "elmo: ESS says abort, so aborting bulk data transfer\n");
      break;
    }
    default:
      printk(KERN_ERR "elmo: unhandled command 0x%02X!\n",
	     elmo_host_data.ataCommandReg & 0xFF);
      save_and_cli(flags);
      ELMO_BRIDGE->error = (ABRT_ERR << 8);
      ELMO_BRIDGE->error = (ABRT_ERR << 8) | 1;
      ELMO_BRIDGE->error = 0;
      elmo_host_data.haveCommand = 0;
      restore_flags(flags);
    }
  }
  elmo_queue_worker_for_interrupt(worker, 
				  ELMO_INT_DSK_RST_REQ | ELMO_INT_DSK_CMD_REQ,
				  ELMO_WORKER_AWAIT_RESET);
  ELMO_BRIDGE->imr |= ELMO_ISR_DSK_RST_REQ | ELMO_ISR_DSK_CMD_REQ;
}

static void elmo_worker_setup_reset(struct elmo_worker *worker)
{
  if (debug) printk(KERN_INFO "elmo: FIFO-flush DMA initiated, wait 5 ms\n");
  elmo_queue_worker_for_timer(worker, 5, ELMO_WORKER_RESET_DONE);
}

static void elmo_worker_reset_done(struct elmo_worker *worker)
{
  if (debug) printk(KERN_INFO "elmo: timer expired, stopping DMA\n");
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
}

static void elmo_worker_service_read_fifo(struct elmo_worker *worker)
{
  int spaceRemaining;
  int took;
  if (!elmo_host_data.pldLoaded) {
    if (debug > 1) printk(KERN_INFO "elmo: not ready, pull worker sleeps\n");
    return;
  }
  elmo_worker_disarm_timer(worker);
  spaceRemaining = (elmo_host_data.llReadBufferTail - elmo_host_data.llReadBufferHead) %
    ELMO_READBUFFER_SIZE;
  if (spaceRemaining == 1) {
    /* the ring buffer is actually full.  Go away until someone drains it */
    if (debug > 1) printk(KERN_INFO "elmo: pull buffer is full, worker sleeps\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_READBUFFER_SPACE;
    return;
  }
  if (spaceRemaining == 0) {
    spaceRemaining = ELMO_READBUFFER_SIZE;
  }
  if (debug > 1) printk(KERN_INFO "elmo: pull buffer has %d bytes available\n",
			spaceRemaining);
  took = 0;
  while (spaceRemaining > 1 && (ELMO_BRIDGE->fsr & ELMO_FIFO_ESS_TO_EBI_NE)) {
      if (0 == useshortword ) {
        elmo_host_data.llReadBuffer[elmo_host_data.llReadBufferHead] =
          (unsigned char) *(ELMO_LOW_LATENCY_FIFO);
        elmo_host_data.llReadBufferHead = (elmo_host_data.llReadBufferHead + 1) %
          ELMO_READBUFFER_SIZE;
        took ++;
        spaceRemaining --;
      }else {
          unsigned short *shortReadBuffer = (unsigned short *)&(elmo_host_data.llReadBuffer[elmo_host_data.llReadBufferHead]);
          *shortReadBuffer = (unsigned short) *(ELMO_LOW_LATENCY_FIFO);
          elmo_host_data.llReadBufferHead = 
              (elmo_host_data.llReadBufferHead + 2) % ELMO_READBUFFER_SIZE;
        took +=2;
        spaceRemaining -=2;
      }

  }
  PLOG(
      PLOG_ELMO_RPC_NEW_DATA_FROM_ESS,
      sizeof(took),
      &took
      );
  PLOG(
      PLOG_ELMO_RPC_NEW_DATA_WAITING_FOR,
      sizeof(spaceRemaining),
      &spaceRemaining
      );

  if (debug > 1) printk(KERN_INFO "elmo: pulled %d bytes to buffer\n",
			took);

#if 0
  /*
    If we didn't get anything this time, arrange to be interrupted
    instantly whenever anything appears in the FIFO.

    Otherwise, arrange to be interrupted if the FIFO reaches the
    partially-full watermark, or after a few ticks.  Also, release the
    data-is-available resource so that the reader (if any) can awaken.
  */
  if (took == 0) {
    if (debug > 1) printk(KERN_INFO "elmo: pullworker sleeps awaiting data\n");
    elmo_queue_worker_for_interrupt(worker, ELMO_INT_ESS_TO_EBI_NE,
				    ELMO_WORKER_SERVICE_READ_FIFO);
    ELMO_BRIDGE->fmr = (ELMO_BRIDGE->fmr & ~ ELMO_FIFO_ESS_TO_EBI_HWM) |
      ELMO_FIFO_ESS_TO_EBI_NE;
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: pull worker sleeps on watermark\n");
    elmo_queue_worker_for_interrupt(worker, ELMO_INT_ESS_TO_EBI_HWM,
				    ELMO_WORKER_SERVICE_READ_FIFO);
    elmo_queue_worker_for_timer(worker, 5, ELMO_WORKER_SERVICE_READ_FIFO);
    ELMO_BRIDGE->fmr = (ELMO_BRIDGE->fmr & ~ ELMO_FIFO_ESS_TO_EBI_NE) |
      ELMO_FIFO_ESS_TO_EBI_HWM;
    elmo_release_chip_resources(ELMO_RESOURCE_READBUFFER_DATA);
  }

#else

  /*
    The logic above makes things very slooow. Scanning directory takes 
    38ms per directory entry, for example.

    Instead, we will wait for NE interrupt always, even if we copy some data.
    It seems the above logic is there to protect us from too much data and 
    too many interrupts. Trouble with that logic is that we could read some,
    and then wait for HWM or the 5 timer ticks, and if the HWM never hits, 
    we suffer 5 timer ticks.
  */
  if (debug > 1) printk(KERN_INFO "elmo: pullworker sleeps awaiting data\n");
  elmo_queue_worker_for_interrupt(worker, ELMO_INT_ESS_TO_EBI_NE,
                                  ELMO_WORKER_SERVICE_READ_FIFO);
  ELMO_BRIDGE->fmr = (ELMO_BRIDGE->fmr & ~ ELMO_FIFO_ESS_TO_EBI_HWM) |
      ELMO_FIFO_ESS_TO_EBI_NE;
  if ( took > 0 ) {
      elmo_release_chip_resources(ELMO_RESOURCE_READBUFFER_DATA);
  }

#endif

}

static void elmo_worker_service_write_fifo(struct elmo_worker *worker)
{
  int dataAvailable;
  int took;
  if (!elmo_host_data.pldLoaded) {
    if (debug > 1) printk(KERN_INFO "elmo: not ready, push worker sleeps\n");
    return;
  }
  dataAvailable = (elmo_host_data.llWriteBufferTail - elmo_host_data.llWriteBufferHead) %
    ELMO_WRITEBUFFER_SIZE;
  if (dataAvailable == 0) {
    if (debug > 1) printk(KERN_INFO "elmo: push buffer is empty, worker sleeps\n");
    /* the ring buffer is empty.  Go away until someone fills it*/
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITEBUFFER_DATA;
    return;
  }
  if (debug > 1) printk(KERN_INFO "elmo: push buffer has %d bytes in it\n",
			dataAvailable);
  took = 0;
  elmo_worker_disarm_deadman(worker);
  worker->waiting_for_ll_fifo = 0;
  while (dataAvailable > 0 && !(ELMO_BRIDGE->fsr & ELMO_FIFO_EBI_TO_ESS_FF)) {
      if (0 == useshortword) {
        *(ELMO_LOW_LATENCY_FIFO) = 
          (unsigned short) elmo_host_data.llWriteBuffer[elmo_host_data.llWriteBufferHead];
        elmo_host_data.llWriteBufferHead = (elmo_host_data.llWriteBufferHead + 1) %
          ELMO_WRITEBUFFER_SIZE;
        dataAvailable --;
        took ++;
      }else {
        *(ELMO_LOW_LATENCY_FIFO) = 
          *((unsigned short *) &(elmo_host_data.llWriteBuffer[elmo_host_data.llWriteBufferHead]));
        elmo_host_data.llWriteBufferHead = (elmo_host_data.llWriteBufferHead + 2) %
          ELMO_WRITEBUFFER_SIZE;
        dataAvailable -=2;
        took +=2;
      }
  }

  PLOG( PLOG_ELMO_RPC_DATA_SENT_TO_FIFO,
        sizeof(dataAvailable),
        &dataAvailable
      );

  if (debug > 1) printk(KERN_INFO "elmo: pushed %d bytes to FIFO\n", took);
  /*
    If there's any data remaining to be sent, arrange to be interrupted
    when the amount of data in the FIFO falls below the watermark.

    Otherwise, arrange to be awakened when somebody puts more data in
    our buffer... and, also, send the "data space is available" signal
    now to awaken any sleeping writers.
  */
  if (dataAvailable > 0) {
    if (debug > 1) printk(KERN_INFO "elmo: push worker sleeps until FIFO below waterline\n");
    /* set a hard limit of 30 seconds. If ESS does not empty the FIFO by that
     * time, we are in trouble and need to reboot the box to recover.
     */
    elmo_worker_set_deadman (worker, 30*HZ);
    worker->waiting_for_ll_fifo = 1;
    elmo_queue_worker_for_interrupt(worker, ELMO_INT_EBI_TO_ESS_LWM,
				    ELMO_WORKER_SERVICE_WRITE_FIFO);
    ELMO_BRIDGE->fmr = ELMO_BRIDGE->fmr | ELMO_FIFO_EBI_TO_ESS_LWM;
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: push worker sleeps until data appears\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITEBUFFER_DATA;
    elmo_release_chip_resources(ELMO_RESOURCE_WRITEBUFFER_SPACE);
  }
}

static void elmo_worker_gate_writer(struct elmo_worker *worker)
{
  if (!(elmo_host_data.resourcesBusy & ELMO_RESOURCE_WRITER_GATE)) {
    worker->elmoResourcesUsed |= ELMO_RESOURCE_WRITER_GATE;
    elmo_host_data.resourcesBusy |= ELMO_RESOURCE_WRITER_GATE;
    elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  } else {
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITER_GATE;
  }
}

static void elmo_worker_become_writer(struct elmo_worker *worker)
{
  if (!(elmo_host_data.resourcesBusy & ELMO_RESOURCE_WRITER)) {
    worker->elmoResourcesUsed |= ELMO_RESOURCE_WRITER;
    elmo_host_data.resourcesBusy |= ELMO_RESOURCE_WRITER;
    elmo_worker_next_state(worker, ELMO_WORKER_WRITER);
  } else {
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITER;
  }
}

static void elmo_worker_writer(struct elmo_worker *worker)
{
  int dataAvailable, spaceAvailable;
  int i;
  dataAvailable = (elmo_host_data.llWriteBufferTail - elmo_host_data.llWriteBufferHead) %
    ELMO_WRITEBUFFER_SIZE;
  spaceAvailable = ELMO_WRITEBUFFER_SIZE - dataAvailable - 1;
  if (spaceAvailable < worker->dataBufferSize) {
    /* Insufficient space.  Go away until the buffer is drained*/
    if (debug > 1) printk(KERN_INFO "elmo: push buffer full, worker sleeps\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_WRITEBUFFER_SPACE;
    return;
  }
  for (i = 0; i < worker->dataBufferSize; i++) {
    elmo_host_data.llWriteBuffer[elmo_host_data.llWriteBufferTail] = worker->dataBuffer[i];
    elmo_host_data.llWriteBufferTail = (elmo_host_data.llWriteBufferTail + 1) %
      ELMO_WRITEBUFFER_SIZE;
  }
  if (debug > 1) printk(KERN_INFO "elmo: pushed %d bytes to ring buffer\n", i);
  /*
    Kick the write servicer
  */
  elmo_release_chip_resources(ELMO_RESOURCE_WRITEBUFFER_DATA);
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
}

static void elmo_worker_become_reader(struct elmo_worker *worker)
{
  if (!(elmo_host_data.resourcesBusy & ELMO_RESOURCE_READER)) {
    if (debug > 1) printk(KERN_INFO "elmo: worker %d becomes the reader\n",
			  worker->sn);
    worker->elmoResourcesUsed |= ELMO_RESOURCE_READER;
    elmo_host_data.resourcesBusy |= ELMO_RESOURCE_READER;
    elmo_worker_next_state(worker, ELMO_WORKER_READER);
  } else {
    if (debug > 1) printk(KERN_INFO "elmo: worker %d waiting to become the reader\n",
			  worker->sn);
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_READER;
  }
}

static void elmo_worker_reader(struct elmo_worker *worker)
{
  int dataAvailable;
  int i;
  dataAvailable = (elmo_host_data.llReadBufferHead - elmo_host_data.llReadBufferTail) %
    ELMO_WRITEBUFFER_SIZE;
  if (dataAvailable == 0) {
    /* No data.  Go away until the read servicer provides some */
    if (debug > 1) printk(KERN_INFO "elmo: reader sleeping to await data\n");
    worker->elmoResourcesNeeded |= ELMO_RESOURCE_READBUFFER_DATA;
    return;
  }
  if (debug > 1) printk(KERN_INFO "elmo: reader finds %d avail, %d wanted\n",
			dataAvailable, worker->dataBufferSize);
  for (i = 0; dataAvailable > 0 && i < worker->dataBufferSize; i++) {
    worker->dataBuffer[i] = elmo_host_data.llReadBuffer[elmo_host_data.llReadBufferTail];
    elmo_host_data.llReadBufferTail = (elmo_host_data.llReadBufferTail + 1) %
      ELMO_WRITEBUFFER_SIZE;
    dataAvailable --;
    worker->dataBufferCount ++;
  }
  /*
    Wake up the process which submitted us
  */
  worker->readCompleted = 1;
  wake_up_interruptible(&readerWaitQueue);
  /*
    Kick the read servicer
  */
  elmo_release_chip_resources(ELMO_RESOURCE_READBUFFER_SPACE);
  /*
    Go into the post-read idle state until the process which submitted
    us grabs the data and marks us as being done.
  */
  elmo_worker_next_state(worker, ELMO_WORKER_READER_POST);
}

static void elmo_worker_reader_post(struct elmo_worker *worker)
{
  /* This is a null/idling state.  Do nothing, go back to sleep. */
  if (debug > 1) printk(KERN_INFO "elmo: worker %d awaiting postprocessing\n",
			worker->sn);
}

static void elmo_queue_worker_for_interrupt (struct elmo_worker *worker,
					  unsigned long interruptsNeeded,
					  enum elmo_worker_state nextState)
{
  unsigned long flags;
  save_and_cli(flags);
  if (debug > 1 && worker->elmoInterruptsPending & ~interruptsNeeded) {
    printk(KERN_DEBUG "elmo: worker %d new state %d old int %08lX add int %08lX\n",
	   worker->sn, nextState, worker->elmoInterruptsPending,
	   interruptsNeeded);
  }
  worker->elmoInterruptsPending |= interruptsNeeded;
  worker->state = nextState;
  restore_flags(flags);
}

static void elmo_queue_worker_for_timer (struct elmo_worker *worker,
				       int timerTicks,
				       enum elmo_worker_state nextState)
{
  unsigned long flags;
  save_and_cli(flags);
  worker->elmoInterruptsPending |= ELMO_INT_TIMER;
  worker->timeout = timerTicks ? timerTicks : 1;
  worker->state = nextState;
  restore_flags(flags);
  queue_task(&elmo_timer_task_struct, &tq_timer);
}

static void elmo_worker_disarm_timer (struct elmo_worker *worker) {
  unsigned long flags;
  save_and_cli(flags);
  worker->elmoInterruptsPending &= ~ELMO_INT_TIMER;
  worker->timeout = 0;
  restore_flags(flags);
}

static void elmo_worker_set_deadman (struct elmo_worker *worker,
				     unsigned long timerTicks)
{
  unsigned long flags;
  save_and_cli(flags);
  worker->deadman = timerTicks ? timerTicks : 20*HZ;
  restore_flags(flags);
  queue_task(&elmo_timer_task_struct, &tq_timer);
}

static void elmo_worker_disarm_deadman (struct elmo_worker *worker)
{
  worker->deadman = 0;
}

static unsigned long elmo_interrupt_for_workers(unsigned long interruptsFired)
{
  unsigned long flags;
  unsigned long intersection;
  unsigned long interruptsTaken;
  struct list_head *this;
  struct elmo_worker *worker;

  save_and_cli(flags);
  interruptsTaken = 0;
  if (debug) printk(KERN_INFO "elmo: process virtual interrupt 0x%08lX\n",
	 interruptsFired);
  list_for_each(this,&elmo_host_data.workers) {
    worker = (struct elmo_worker *) this;
    if (worker->elmoInterruptsPending) {
      if (debug) printk(KERN_INFO "elmo: worker %d pending on 0x%08lX\n",
			worker->sn, worker->elmoInterruptsPending);
      intersection = worker->elmoInterruptsPending & interruptsFired;
      if (intersection) {
	interruptsTaken |= intersection;
	worker->elmoInterruptsFired = intersection;
	worker->elmoInterruptsPending &= ~intersection;
	worker->go = 1;
	elmo_worker_disarm_timer(worker);
	if (debug) printk(KERN_INFO "elmo: woke up the worker\n");
      }
    }
  }
  return interruptsTaken;
}  

/*
 *	elmo_run_workers - run the individual state machines for the
 *	ATAPI requests on the queue.
 *
 */

static int elmo_run_workers()
{
  int done = 1;
  struct list_head *this, *next;
  struct elmo_worker *worker;

  if (!elmo_host_data.pldLoaded) {
    return done;
  }

  if (debug > 2) printk(KERN_INFO "elmo: running worker queue\n");

  list_for_each_safe(this,next,&elmo_host_data.workers) {
    worker = (struct elmo_worker *) this;
    if (worker->go) {
      if (debug > 2) printk(KERN_INFO "elmo: worker %d go!\n", worker->sn);
      worker->go = 0;

      PLOG( PLOG_ELMO_RPC_WORKER_STATE, sizeof(worker->state), &worker->state );

      switch (worker->state) {
      case ELMO_WORKER_SETUP:
	elmo_worker_setup(worker);
	break;
      case ELMO_WORKER_GET_BULK_INTERFACE:
	elmo_worker_get_bulk_interface(worker);
	break;
      case ELMO_WORKER_GET_ESS_COMMAND:
	elmo_worker_get_ess_command(worker);
	break;
      case ELMO_WORKER_SETUP_DMA:
	elmo_worker_setup_dma(worker);
	break;
      case ELMO_WORKER_AWAIT_FIFO:
	elmo_worker_await_fifo(worker);
	break;
      case ELMO_WORKER_FINISH_DMA:
	elmo_worker_finish_dma(worker);
	break;
      case ELMO_WORKER_DONE:
	elmo_worker_done(worker);
	list_del((struct list_head *) worker);
	kfree(worker);
	worker = NULL;
	break;
      case ELMO_WORKER_AWAIT_RESET:
	elmo_worker_await_reset(worker);
	break;
      case ELMO_WORKER_SETUP_RESET:
	elmo_worker_setup_reset(worker);
	break;
      case ELMO_WORKER_RESET_DONE:
	elmo_worker_reset_done(worker);
	break;
      case ELMO_WORKER_SERVICE_READ_FIFO:
	elmo_worker_service_read_fifo(worker);
	break;
      case ELMO_WORKER_SERVICE_WRITE_FIFO:
	elmo_worker_service_write_fifo(worker);
	break;
      case ELMO_WORKER_GATE_WRITER:
	elmo_worker_gate_writer(worker);
	break;
      case ELMO_WORKER_BECOME_WRITER:
	elmo_worker_become_writer(worker);
	break;
      case ELMO_WORKER_WRITER:
	elmo_worker_writer(worker);
	break;
      case ELMO_WORKER_BECOME_READER:
	elmo_worker_become_reader(worker);
	break;
      case ELMO_WORKER_READER:
	elmo_worker_reader(worker);
	break;
      case ELMO_WORKER_READER_POST:
	elmo_worker_reader_post(worker);
	break;
      default:
	break;
      }
    } else {
      if (debug > 2) printk(KERN_INFO "elmo: worker %d nogo\n", worker->sn);
    }
  }
  list_for_each(this,&elmo_host_data.workers) {
    worker = (struct elmo_worker *) this;
    done &= ! worker->go;
  }
  if (debug > 2) printk(KERN_INFO "elmo: queue runner says %sdone\n",
			done ? "" : "not ");
  return done;
}

/*
  The coroutine code has been gently and lovingly lifted from
  NCR5380 - go see the comments there.
*/

static unsigned long main_running = 0;

static __inline__ void run_main(void)
{
  if (!test_and_set_bit(0, &main_running)) {
    elmo_main();
  }
}

/*
  Coroutine runner task
*/

static void elmo_task (void *whatever)
{
  (void) whatever;
  run_main();
}

/*
  Timer task.  Run through all known workers, and kick awake any of
  them who were waiting for a timer countdown and whose countdown
  has expired.  Also, abort any workers which had a deadman timer
  (these are usually background/buffered write commands
*/

static void elmo_timer_task (void *whatever)
{
    struct list_head *this;
    struct elmo_worker *worker;
    int wakeup = 0;
    int rearm = 0;
    (void) whatever;
    list_for_each(this,&elmo_host_data.workers) {
        worker = (struct elmo_worker *) this;
        if (worker->deadman != 0) {
             worker->deadman --;
             if (worker->deadman == 0) {
                   worker->aborting = 1;
                   worker->go = 1;
                   wakeup = 1;
                   if (worker->request) {
                       worker->request->dio_request->status |= ELMO_ERROR_TIMEOUT;
                   }
                   if(0 != worker->waiting_for_ll_fifo) {
                       worker->go = 0;
                       elmo_host_data.essWedged = 1;
                       worker->waiting_for_ll_fifo = 0;
                       printk(KERN_ERR "ESS wedged! EBI to ESS FIFO full for 30 seconds. Setting system up for reboot\n");
                   }
                   printk(KERN_ERR "elmo: deadman timer aborting worker #%d\n", worker->sn);
             } else {
                   rearm = 1;
             }
        }
        if (worker->timeout > 0 &&
           (worker->elmoInterruptsPending & ELMO_INT_TIMER)) {
             worker->timeout --;
             if (worker->timeout == 0) {
                   worker->elmoInterruptsPending &= ~ELMO_INT_TIMER;
                   worker->elmoInterruptsFired |= ELMO_INT_TIMER;
                   worker->go = 1;
                   if (worker->request) {
                       worker->request->dio_request->status |= ELMO_ERROR_TIMEOUT;
                   }
                   wakeup = 1;
             } else {
                   rearm = 1;
             }
        }
    }
    if (wakeup) {
       queue_task (&elmo_task_struct, &tq_immediate);
       mark_bh (IMMEDIATE_BH);
    }
#ifdef SIMULATE_IRQ
    do_elmo_intr(ELMO_IRQ, NULL, NULL);
    rearm = runIRQs;
#endif
    if (rearm) {
      queue_task(&elmo_timer_task_struct, &tq_timer);
    }
}


/**
 *	elmo_main	-	Elmo state machines
 *
 *	elmo_main is a coroutine that runs as long as more work can 
 *      be done on the Elmo host adapter in a system.  Both 
 *      elmo_queue_command() and elmo_intr() will try to start it 
 *      in case it is not running.
 * 
 *	Locks; The caller must hold the io_request_lock. The lock will still be
 *	held on return but may be dropped while running. Called functions take
 *	the DMA lock.
 */

static void elmo_main(void) {
   int done;
   unsigned int iflags;

   /*
    * We run (with interrupts disabled) until we're sure that none of
    * the host adapters have anything that can be done, at which point 
    * we set main_running to 0 and exit.
    *
    * Interrupts are enabled before doing various other internal 
    * instructions, after we've decided that we need to run through
    * the loop again.
    *
    * this should prevent any race conditions.
    *
    */

   do {
     /* Lock held here */
     done = 1;
     {
       write_lock_irqsave(&elmo_host_data.queueLock, iflags);
       done &= elmo_run_workers();
     }
   } while (!done);
   /* Exit lock held */
   clear_bit(0, &main_running);

}

static void show_elmo_intr(unsigned long combinedISR)
{
  const int numIntrs = 32;
  const char *intrs[] = {
    " EBI_TO_ESS_NE",
    " EBI_TO_ESS_WM",
    " EBI_TO_ESS_FF",
    " BROADCOM_DMA",
    " ESS_TO_EBI_NE",
    " ESS_TO_EBI_WM",
    " TIMER",
    " RFU_BIT_7",
    " EBI_TO_IDE_NE",
    " EBI_TO_IDE_WM",
    " EBI_TO_IDE_FF",
    " RFU_BIT_11",
    " IDE_TO_EBI_NE",
    " IDE_TO_EBI_WM",
    " RFU_BIT_14",
    " RFU_BIT_14",
    " BUS_TIMEOUT",
    " ELMO_ASIC_DMA_RD_DONE",
    " ELMO_ASIC_DMA_WR_DONE",
    " RFU_BIT_3",
    " RFU_BIT_4",
    " DSK_RST_REQ",
    " DSK_CMD_REQ",
    " INT_FROM_ESS",
    " EBI_TO_ESS_OVR",
    " EBI_TO_ESS_UDR",
    " ESS_TO_EBI_OVR",
    " ESS_TO_EBI_UDR",
    " EBI_TO_IDE_OVR",
    " EBI_TO_IDE_UDR",
    " IDE_TO_EBI_OVR",
    " IDE_TO_EBI_UDR" };
  int i;
  if (!combinedISR) {
    return;
  }
  printk(KERN_INFO "elmo: interrupt");
  for (i = numIntrs - 1; i >= 0; i--) {
    if (combinedISR & (1 << i)) {
      printk(intrs[i]);
    }
  }
  printk("\n");
}


static void show_elmo_errors(elmo_error_codes_e errors) 
{
  const char *elmo_err_strs[] = {
      " ELMO_ERROR_HANG",
      " ELMO_ERROR_TIMEOUT",
      " ELMO_ERROR_ABORT",
      " ELMO_ERROR_ATA_ABORT",
      " ELMO_ERROR_UNKNOWN",
      " ELMO_ERROR_SGAT_DIO", 
      " ELMO_ERROR_BULK_WORKER_INIT", 
      " ELMO_ERROR_ATA",
      " ELMO_ERROR_ERROR_END" 
  };
  const int numErrs = sizeof(elmo_err_strs)/sizeof(elmo_err_strs[0]);
  int i = 0;
  for (i = numErrs - 1; i >= 0; i--) {
    if (errors & (1 << i)) {
      printk(elmo_err_strs[i]);
    }
  }
  printk("\n");
}


/**
 *
 *      Broadcom "end of DMA" interrupt handlers
 *
 */

static void broadcom_intr(int irq, int *useFlag, unsigned long softInterrupt)
{
  unsigned long flags;
  unsigned long interruptsTaken;

  save_and_cli(flags);

  if (debug) printk(KERN_INFO "elmo: Broadcom DMA interrupt %d triggered!\n", irq);

  disable_irq(irq);

  interruptsTaken = 0;
  interruptsTaken |= elmo_interrupt_for_workers(softInterrupt);

  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);

  if (interruptsTaken) {
      *useFlag = 0;
  }

  restore_flags(flags);
  
  if (!interruptsTaken) {
    printk(KERN_ERR "Oops!  No worker took Broadcom DMA interrupt %d!\n", irq);
  }
}

static void broadcom_isb_intr(int irq, void *dev_id, struct pt_regs *regs)
{
  (void) dev_id;
  (void) regs;
  *BCM_DMA_CH1_INT_STAT = 0x4;
  *BCM_DMA_CH2_INT_STAT = 0x4;
  barrier();
  broadcom_intr(irq, &elmo_broadcom_isb_enabled, ELMO_INT_ISB_DMA);
}

/**
 * 	elmo_intr	-	Elmo interrupt handler
 *
 *	Handle interrupts.
 *
 *	Locks: caller must hold the io_request lock.
 */

static void elmo_intr(int irq, void *dev_id, struct pt_regs *regs) 
{
  unsigned long interruptsTaken;
  unsigned long wakeups;
  unsigned short bridgeISR, bridgeFISR;

  /*
    Do the magic here.  Read out and clear the interrupts in the main ISR
    which triggered us.  Also, read out the FIFO status register
    conditions which triggered us, and mask them off so that they
    do not trigger again unless some worker unmasks them.
  */

  bridgeISR = ELMO_BRIDGE->isr & ELMO_BRIDGE->imr;
  barrier();
  ELMO_BRIDGE->isr = bridgeISR; /* clear everything we can... */
  barrier();

  bridgeFISR = ELMO_BRIDGE->fsr & ELMO_BRIDGE->fmr;
  barrier();
  ELMO_BRIDGE->fmr &= ~bridgeFISR;  /* mask off the things which triggered */
  barrier();

  /*
    Combine the two 16-bit interrupt sets in a single 32-bit set which
    defines the 'virtual' interrupts we need to process.
  */

  wakeups = (((unsigned long) bridgeISR) << 16) | bridgeFISR;

  PLOG( PLOG_ELMO_RPC_ELMO_INTERRUPT, sizeof(wakeups), &wakeups );

  if (debug) {
    show_elmo_intr(wakeups);
  }

  /*
    Kick awake any workers who were waiting for any of the interrupts
    that we just noticed, and then tickle the bottom-half task
  */

  if (wakeups) {

    interruptsTaken = 0;

    interruptsTaken |= elmo_interrupt_for_workers(wakeups);

    if (interruptsTaken & ELMO_INT_DMA_RD_IRQ) {
        elmo_asic_read_dma_enabled = 0;
        ELMO_BRIDGE->dma_control &=~(0x1000U);
    }else if (interruptsTaken & ELMO_INT_DMA_WR_IRQ) {
        elmo_asic_write_dma_enabled = 0;
        ELMO_BRIDGE->dma_control &=~(0x2000U);
    }
    barrier();

    queue_task (&elmo_task_struct, &tq_immediate);
    mark_bh (IMMEDIATE_BH);

    if (interruptsTaken != wakeups) {
      printk(KERN_ERR "elmo: OOPS:  Unclaimed virtual interrupt 0x%08lX\n",
	     wakeups & ~interruptsTaken);
      show_elmo_intr(wakeups & ~interruptsTaken);
    }
  }
	   
}

/**
 *	do_elmo_intr
 *	@irq: interrupt number
 *	@dev_id: device info
 *	@regs: registers (unused)
 *
 *	Takes the io_request_lock and invokes the Elmo interrupt
 *	handler code
 *
 *	Locks: takes and releases the io_request lock
 */

static void do_elmo_intr(int irq, void *dev_id, struct pt_regs *regs) {
  unsigned long flags;

  volatile unsigned short imr = 0;
  volatile unsigned short fmr = 0;

  spin_lock_irqsave(&io_request_lock, flags);
  elmo_intr(irq, dev_id, regs);
  spin_unlock_irqrestore(&io_request_lock, flags);

  /* clear the interrupt mask register and set it to the same value. We need to
   * do this because the 7317 does not support level triggered interrupts on
   * GPIOs. We have to clear and set the ASIC interrupt mask register to force
   * an edge transition on the GPIO line, if an interrupt is pending. This will
   * hold the interrupt line active, so the 7317 will detect the interrupt and
   * service it. Otherwise, there is a possibility for a pending interrupt to
   * be missed
   */
  imr = ELMO_BRIDGE->imr;
  barrier();
  ELMO_BRIDGE->imr = 0;
  barrier();
  ELMO_BRIDGE->imr = imr;

  /* clear the fifo interrupt mask register and set it to the same value. 
   * We need to do this because the 7317 does not support level triggered 
   * interrupts on GPIOs. We have to clear and set the ASIC interrupt mask 
   * register to force an edge transition on the GPIO line, if an interrupt
   * is pending. This will hold the interrupt line active, so the 7317 will 
   * detect the interrupt and service it. Otherwise, there is a possibility 
   * for a pending interrupt to be missed
   */
  fmr = ELMO_BRIDGE->fmr;
  barrier();
  ELMO_BRIDGE->fmr = 0;
  barrier();
  ELMO_BRIDGE->fmr = fmr;
  barrier();
}

static void elmo_dump_bridge_regs()
{
    printk(KERN_ERR "elmo:  ctl %04X\n", ELMO_BRIDGE->control);
    printk(KERN_ERR "elmo:  isr %04X\n", ELMO_BRIDGE->isr);
    printk(KERN_ERR "elmo:  imr %04X\n", ELMO_BRIDGE->imr);
    printk(KERN_ERR "elmo:  fsr %04X\n", ELMO_BRIDGE->fsr);
    printk(KERN_ERR "elmo:  fmr %04X\n", ELMO_BRIDGE->fmr);
    printk(KERN_ERR "elmo:  dma %04X\n", ELMO_BRIDGE->dma_control);
}


#define EBI_CLOCK_27MHZ         27000000
#define EBI_CLOCK_40MHZ         40500000

static unsigned long elmo_get_ebi_clock_speed()
{
      unsigned long ebi_speed = 0;
      struct timeval start, end;
      unsigned short revision = 0;
      int i = 0;

      do_gettimeofday(&start);

      for (i = 0; i < 100000; ++i) {
          revision = ELMO_BRIDGE->revision;
          barrier();
      }

      do_gettimeofday(&end);
      
      ebi_speed = TO_USEC(end) - TO_USEC(start);
      printk(KERN_INFO "ebi speed check result : %ld\n", ebi_speed);
      if (ebi_speed > 44000) {
          ebi_speed = EBI_CLOCK_27MHZ;
      }else {
          ebi_speed = EBI_CLOCK_40MHZ;
      }
      return ebi_speed;
}

#define PRINTP(x) len += sprintf(buffer+len, x)
#define ANDP ,

static int elmo_proc_info(char *buffer, char **start, off_t offset,
			int length, int *eof, void *data)
{
  int len = 0;
  unsigned long flags;
  (void) start;
  (void) offset;
  (void) length;
  (void)eof;
  (void)data;
  if (debug) printk(KERN_DEBUG "In elmo_proc_info()\n");
  save_and_cli(flags);
  if (elmo_host_data.pldLoaded) {
    PRINTP("Bridge : UP, version %02X\n" ANDP (0xFF & ELMO_BRIDGE->revision));
  } else {
    PRINTP("Bridge : DOWN\n");
  }
  *start = buffer + offset;
  len -= offset;
  if (len > length)
    len = length;
  restore_flags(flags);
  return len;
}

/*
 * Handlers for read/write/ioctl on the Elmo low-latency interface.
 */

ssize_t elmo_ll_read (struct file *filp, char *b, size_t t, loff_t *l)
{
  struct elmo_worker *worker;
  unsigned int flags;
  int ret;
  if (0 != elmo_host_data.essWedged && !noesswatchdog ) {
      return -ENOLINK;
  }
  if (t == 0) {
    return 0;
  } else if (t > ELMO_READBUFFER_SIZE) {
    t = ELMO_READBUFFER_SIZE;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_KERNEL);
  if (debug > 1) printk(KERN_INFO "elmo: create worker for reading\n");
  if (!worker) {
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++elmo_host_data.workerSN;
  if (debug > 1) printk(KERN_INFO "elmo: allocating %d-byte buffer\n", t);
  worker->dataBuffer = kmalloc(t, GFP_KERNEL);
  if (!worker->dataBuffer) {
    kfree(worker);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  if (debug > 1) printk(KERN_INFO "elmo: queueing worker %d for read\n",
			worker->sn);
  elmo_worker_next_state(worker, ELMO_WORKER_BECOME_READER);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker, &elmo_host_data.workers);
  restore_flags(flags);
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  if (debug > 1) printk(KERN_INFO "elmo: waiting on worker\n");
  ret = wait_event_interruptible(readerWaitQueue, worker->readCompleted);
  if (ret == 0) {
    copy_to_user (b, worker->dataBuffer, worker->dataBufferCount);
    ret = worker->dataBufferCount;
  }
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  return ret;
}

ssize_t elmo_ll_readv (struct file *filp, const struct iovec *iov, 
		       unsigned long count, loff_t *l)
{
  struct elmo_worker *worker;
  unsigned int flags;
  size_t t;
  size_t have, offset, take;
  int i;
  int ret = 0;
  t = 0;
  if (0 != elmo_host_data.essWedged && !noesswatchdog ) {
      return -ENOLINK;
  }
  if (debug > 1) printk(KERN_INFO "elmo: summing iovec for readv\n");
  for (i = 0; i < count; i++) {
    if (verify_area(VERIFY_WRITE, iov[i].iov_base, iov[i].iov_len)) {
      return -EFAULT;
    }
    t += iov[i].iov_len;
  }
  if (debug > 1) printk(KERN_INFO "elmo: readv total size is %d\n", t);
  if (t == 0) {
    return 0;
  } else if (t > ELMO_READBUFFER_SIZE) {
    t = ELMO_READBUFFER_SIZE;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_KERNEL);
  if (!worker) {
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++elmo_host_data.workerSN;
  worker->dataBuffer = kmalloc(t, GFP_KERNEL);
  if (!worker->dataBuffer) {
    kfree(worker);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  elmo_worker_next_state(worker, ELMO_WORKER_BECOME_READER);
  save_and_cli(flags);
  list_add_tail((struct list_head *) worker, &elmo_host_data.workers);
  restore_flags(flags);

  PLOG( PLOG_ELMO_RPC_READV_STARTED, 0, 0 );

  run_main();
  have = worker->dataBufferCount;

  PLOG( PLOG_ELMO_RPC_READV_HAS, sizeof( have ), &have );

  if ( have == 0 ) {
      queue_task (&elmo_task_struct, &tq_immediate);
      mark_bh (IMMEDIATE_BH);
      ret = wait_event_interruptible(readerWaitQueue, worker->readCompleted);
  }
  if (debug > 1) printk(KERN_INFO "readv back from wait, code %d\n", ret);
  if (ret != 0) {
    elmo_worker_next_state(worker, ELMO_WORKER_DONE);
    queue_task (&elmo_task_struct, &tq_immediate);
    mark_bh (IMMEDIATE_BH);
    return ret;
  }
  offset = 0;
  have = worker->dataBufferCount;

  PLOG( PLOG_ELMO_RPC_READV_HAS2, sizeof( have ), &have );

  if (debug > 1) printk(KERN_INFO "readv has %d available\n", have);
  for (i = 0; i < count && have > 0; i++) {
    take = have;
    if (take > iov[i].iov_len) {
      take = iov[i].iov_len;
    }
    copy_to_user(iov[i].iov_base, &worker->dataBuffer[offset], take);
    offset += take;
    have -= take;
  }
  elmo_worker_next_state(worker, ELMO_WORKER_DONE);
  queue_task (&elmo_task_struct, &tq_immediate);
  mark_bh (IMMEDIATE_BH);
  return offset;
}

ssize_t elmo_ll_write_common (struct file *filp, const char *b, size_t t,
			      loff_t *l, int defer)
{
  struct elmo_worker *worker;
  unsigned int flags;
  int ret;
  if (0 != elmo_host_data.essWedged && !noesswatchdog ) {
      return -ENOLINK;
  }
  if (t == 0) {
    return 0;
  } else if (t > 4095) {
    return -EINVAL;
  }
  if (debug > 1) printk(KERN_INFO "elmo: writer sem now %d\n",
		    sem_getcount(&elmo_host_data.writeSem));
  ret = down_interruptible(&elmo_host_data.writeSem);
  if (ret) {
    return ret;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_KERNEL);
  if (!worker) {
    up(&elmo_host_data.writeSem);
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++elmo_host_data.workerSN;
  worker->holdingSemaphore = 1;
  worker->dataBuffer = kmalloc(t, GFP_KERNEL);
  if (debug > 1) printk(KERN_INFO "elmo: worker %d downs sem to %d\n", worker->sn,
		    sem_getcount(&elmo_host_data.writeSem));
  if (!worker->dataBuffer) {
    kfree(worker);
    up(&elmo_host_data.writeSem);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  copy_from_user (worker->dataBuffer, b, t);
  save_and_cli(flags);
  if (defer) {
    elmo_worker_next_state(worker, ELMO_WORKER_GATE_WRITER);
    elmo_host_data.resourcesBusy |= ELMO_RESOURCE_WRITER_GATE;
  } else {
    elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  }
  list_add_tail((struct list_head *) worker, &elmo_host_data.workers);
  if (!defer) {
    queue_task (&elmo_task_struct, &tq_immediate);
    mark_bh (IMMEDIATE_BH);
  }
  restore_flags(flags);
  return t;
}

ssize_t elmo_ll_write (struct file *filp, const char *b, size_t t, loff_t *l)
{
  return elmo_ll_write_common (filp, b, t, l, 0);
}

ssize_t elmo_ll_write_deferred (struct file *filp, const char *b, size_t t,
				loff_t *l)
{
  return elmo_ll_write_common (filp, b, t, l, 1);
}

ssize_t elmo_ll_writev_common (struct file *filp, const struct iovec *iov, 
			unsigned long count, loff_t *l, int defer)
{
  struct elmo_worker *worker;
  size_t t;
  size_t offset, take;
  int i;
  int take_count = count;
  unsigned int flags;
  int ret;
  t = 0;
  if (0 != elmo_host_data.essWedged && !noesswatchdog ) {
      return -ENOLINK;
  }
  for (i = 0; i < count; i++) {
    if (t + iov[i].iov_len > 4095) {
        take_count = i;
        break;
    }
    t += iov[i].iov_len;
  }
  if (t == 0) {
    return 0;
  } else if (t > 4095) {
    return -EINVAL;
  }
  if (debug > 1) printk(KERN_INFO "elmo: writer sem now %d\n",
		    sem_getcount(&elmo_host_data.writeSem));
  ret = down_interruptible(&elmo_host_data.writeSem);
  if (ret) {
    return ret;
  }
  worker = (struct elmo_worker *) kmalloc(sizeof (struct elmo_worker),
					  GFP_KERNEL);
  if (!worker) {
    up(&elmo_host_data.writeSem);
    return -ENOMEM;
  }
  memset(worker, 0, sizeof *worker);
  worker->sn = ++elmo_host_data.workerSN;
  worker->holdingSemaphore = 1;
  if (debug > 1) printk(KERN_INFO "elmo: worker %d downs sem to %d\n", worker->sn,
		    sem_getcount(&elmo_host_data.writeSem));
  worker->dataBuffer = kmalloc(t, GFP_KERNEL);
  if (!worker->dataBuffer) {
    kfree(worker);
    up(&elmo_host_data.writeSem);
    return -ENOMEM;
  }
  worker->dataBufferSize = t;
  offset = 0;
  for (i = 0; i < take_count; i++) {
    take = iov[i].iov_len;
    copy_from_user(&worker->dataBuffer[offset], iov[i].iov_base, take);
    offset += take;
  }
  elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  save_and_cli(flags);
  if (defer) {
    elmo_worker_next_state(worker, ELMO_WORKER_GATE_WRITER);
    elmo_host_data.resourcesBusy |= ELMO_RESOURCE_WRITER_GATE;
  } else {
    elmo_worker_next_state(worker, ELMO_WORKER_BECOME_WRITER);
  }
  list_add_tail((struct list_head *) worker, &elmo_host_data.workers);
  restore_flags(flags);

  run_main();

  save_and_cli(flags);
  if (!defer) {
      queue_task (&elmo_task_struct, &tq_immediate);
      mark_bh (IMMEDIATE_BH);
  }
  restore_flags(flags);

  PLOG( PLOG_ELMO_RPC_LL_WRITE_DONE, 0, 0);

  return t;
}

ssize_t elmo_ll_writev (struct file *filp, const struct iovec *iov, 
			unsigned long count, loff_t *l) 
{
  return elmo_ll_writev_common (filp, iov, count, l, 0);
}
  

ssize_t elmo_ll_writev_deferred (struct file *filp, const struct iovec *iov, 
				 unsigned long count, loff_t *l) 
{
  return elmo_ll_writev_common (filp, iov, count, l, 1);
}

static int elmo_request_init(struct elmo_request    *elmo_req,
                             Sgat_dio_request   *dio_request)
{
      memset(elmo_req, 0, sizeof(struct elmo_request));
      elmo_req->sg_list   = dio_request->sclp;
      elmo_req->sg_count  = dio_request->sclp_count;
      switch (dio_request->dxfer_dir) {
          case SGAT_DIO_DXFER_FROM_DEV:
              elmo_req->direction = ELMO_BULK_IO_FROM_ESS;
              break;
          case SGAT_DIO_DXFER_TO_DEV:
              elmo_req->direction = ELMO_BULK_IO_TO_ESS;

              if (debug)
                printk(KERN_INFO "elmo: clearing IDE FIFO prior to bulk transfer\n");

              ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
              barrier();
              ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
              barrier();

              break;
          default:
              return -EINVAL;
      }
      elmo_req->xferId    = dio_request->xferId;
      elmo_req->dio_request = dio_request;

      return 0;
}

static int elmo_initiate_bulk_transfer(struct elmo_request *cur_req, 
                                  Sgat_dio_hdr_t *user_dio)
{
      int result = 0;

      if (NULL != elmo_host_data.user_request) {
          printk(KERN_ERR "elmo: attempt to start bulk transfer when a previous call did not cancel/wait for completion\n");
          panic("elmo: Yikes! Attempting to initiate bulk transfer on unclean system\n");

#if 0
          printk(KERN_ERR "elmo: cleaning up abandoned bulk transfer state\n");
          sgat_dio_done_nocopy( elmo_host_data.user_request);
          free_sgat_dio_request(elmo_host_data.user_request);
          elmo_host_data.user_request = NULL;
          elmo_host_data.bulkIOCompleted = 0;
#endif
      }

      if ( 0 != elmo_host_data.bulkIOCompleted) {
          return -EBUSY;
      }



      PLOG(PLOG_ELMO_BULK_TRANSFER_START,sizeof(user_dio->dxfer_len),&user_dio->dxfer_len);

      Sgat_dio_request *dio_request = alloc_sgat_dio_request();
      if (NULL == dio_request) {
          kfree(cur_req);
          return -ENOMEM;
      }

      result = sgat_dio_build_sg_list(user_dio, dio_request);
      if (result) {
          dio_request->status |= ELMO_ERROR_SGAT_DIO;
          unpin_memory(dio_request);
          kfree(cur_req);
          return result;
      }

      result = elmo_request_init(cur_req,dio_request);
      if (result < 0) {
          dio_request->status |= ELMO_ERROR_UNKNOWN;
          unpin_memory(dio_request);
          kfree(cur_req);
          return result;
      }

      PLOG(PLOG_ELMO_BULK_LOG_ID,sizeof(cur_req->xferId), &cur_req->xferId);
      struct elmo_worker *worker = elmo_bulk_io_init(cur_req);
      if (NULL == worker) {
          dio_request->status |= ELMO_ERROR_BULK_WORKER_INIT;
          unpin_memory(dio_request);
          kfree(cur_req);
          return result;
      }

      elmo_worker_set_deadman (worker, (user_dio->timeout*HZ)/1000);

      queue_task (&elmo_task_struct, &tq_immediate);
      mark_bh (IMMEDIATE_BH);

      return 0;
}

static int elmo_complete_bulk_transfer() 
{
    int ret = 0;
      if (debug > 1) printk(KERN_INFO "elmo: waiting on worker\n");
      wait_event(bulkIoWaitQueue, elmo_host_data.bulkIOCompleted);
      elmo_host_data.bulkIOCompleted = 0;
      if (NULL != elmo_host_data.user_request) {
           if (elmo_host_data.user_request->status) {
               elmo_error_codes_e e = 
                   (elmo_error_codes_e)elmo_host_data.user_request->status;

               /* For normal (non debug) operation, we dont have to report the 
                * error string to the app nor do we have to print it. 
                * We leave it in, for debug runs, however, to make sure that
                * the FIFOs are full.
                */
               if (!debug) {
                   e &= ~ELMO_ERROR_ATA_ABORT;
               }

               if (e) {
                 printk(KERN_ERR "elmo: bulk transfer failed : ");
                 show_elmo_errors(
                     (elmo_error_codes_e)elmo_host_data.user_request->status
                                  );
                 elmo_dump_bridge_regs();
               }
               ret = -EIO;
           }

           unpin_memory(elmo_host_data.user_request); /* user_request is freed here */
           elmo_host_data.user_request    = NULL;
      }
      barrier();

      return ret;
}

      

int elmo_ll_ioctl (struct inode *i, struct file *f, unsigned int cmd,
		   unsigned long arg)
{
  int result = 0;
  unsigned long flags;

  if (0 != elmo_host_data.essWedged && !noesswatchdog ) {
      return -ENOLINK;
  }

  if (cmd != kElmoDownloadPLD && cmd != kElmoSetDebugMode &&
      cmd != kElmoHasTei && !elmo_host_data.pldLoaded) {
    printk(KERN_INFO "elmo: access ioctl() on a dead bridge!\n");
    return -EINVAL;
  }
    
  switch( cmd )
    {
    case kElmoReadBridge:
      {
	ElmoIODescriptor descriptor;

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );

	{
	  short *dest = (short *)(descriptor.data);
	  volatile short *src = GetELMOBridgeAddress( descriptor.address );
	  int i;

	  for( i = 0; i < descriptor.length; i += 2 )
	    {
	      short value = *src;
	      barrier();
	      put_user( value, dest );
	      if (debug > 1) {
		printk(KERN_INFO "elmo: peek bridge %08lX has  %04X\n",
		       (unsigned long) src, value);
	      }
	      ++src; ++dest;
	    }
	}
	break;
      }
    case kElmoWriteBridge:
      {
	ElmoIODescriptor descriptor;

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );

	{
	  short *src = (short *)(descriptor.data);
	  volatile short *dest = GetELMOBridgeAddress( descriptor.address );
	  int i;

	  for( i = 0; i < descriptor.length; i += 2 )
	    {
	      short value;
	      get_user( value, src );
	      if (debug > 1) {
		printk(KERN_INFO "elmo: poke bridge %08lX with %04X\n",
		       (unsigned long) dest, value);
	      }
	      *dest = value;
	      barrier();
	      ++src; ++dest;
	    }
	}
	break;
      }

    case kElmoWriteBridgeMasked:
      {
	ElmoMaskDescriptor descriptor;
	volatile short *dest;

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );
	dest = GetELMOBridgeAddress( descriptor.address );

	save_and_cli(flags);

	*dest = (*dest & ~descriptor.mask) |
	  (descriptor.bits & descriptor.mask);

	restore_flags(flags);

	break;
      }

    case kElmoDownloadPLD:
      {
	ElmoPLDDescriptor descriptor;
	int i;
	
	if (elmo_host_data.pldLoaded) {
	  printk(KERN_INFO "elmo: WARNING - reprogramming a live bridge!\n");
          elmo_host_data.pldLoaded = 0;
#ifdef SIMULATE_IRQ
	  runIRQs = 0;
#endif
	}

	copy_from_user( &descriptor, (void*)arg, sizeof( descriptor ) );

	{

	  unsigned char *src = (unsigned char *)(descriptor.data);

	  volatile unsigned char *dest = (unsigned char *) ELMO_BRIDGE;

	  // PROGRAM is connected to GPIO 33
	  // WRITE signal is connected to GPIO 3
        
	  WriteRegister( BCM_GPIO_SEL0, 
			 ReadRegister( BCM_GPIO_SEL0, 1 ) & ~0x30U,
			 1 );

	  // Set both signals off (GPIO 1)
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) | 0x8,
			 1 );
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) | 0x2,
			 1 );

	  // output rather than open drain output
	  WriteRegister( BCM_GPIO_BASE+8, 
			 ReadRegister( BCM_GPIO_BASE+8, 1 ) & ~0x8U,
			 1 );
	  // output rather than input
	  WriteRegister( BCM_GPIO_BASE+7, 
			 ReadRegister( BCM_GPIO_BASE+7, 1 ) & ~0x8U,
			 1 );
                       
	  // output rather than open drain output
	  WriteRegister( BCM_GPIO_BASE+0x55, 
			 ReadRegister( BCM_GPIO_BASE+0x55, 1 ) & ~0x2U,
			 1 );
	  // output rather than input
	  WriteRegister( BCM_GPIO_BASE+0x53, 
			 ReadRegister( BCM_GPIO_BASE+0x53, 1 ) & ~0x2U,
			 1 );

	  // Set both signals off (GPIO 1)
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) | 0x8,
			 1 );
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) | 0x2,
			 1 );

	  // Pulse the "Program" signal to the FPGA low and then high. 
        
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) & ~0x2U,
			 1 );
	  elmo_wait(1);
	  WriteRegister( BCM_GPIO_BASE+0x51,
			 ReadRegister( BCM_GPIO_BASE+0x51, 1 ) | 0x2U,
			 1 );

	  // Set the "Write" signal to the FPGA low. 
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) & ~0x8U,
			 1 );

	  printk(KERN_INFO "elmo: loading %ld bytes from 0x%p\n",
		 descriptor.length, src);

	  for( i = 0; i < descriptor.length; ++i )
	    {
	      unsigned char value;
	      get_user( value, src );
	      *dest = value;
	      ++src;
	    }

	  elmo_wait ( 1 );

	  /* take FPGA out of programming mode */
	  WriteRegister( BCM_GPIO_BASE+3,
			 ReadRegister( BCM_GPIO_BASE+3, 1 ) | 0x8,
			 1 );

	  /* make CS3 be a 16 bit device again */
          *BCM_EBI_CS3_CTRL = 0x08004291U;
          ELMO_BRIDGE->ide_waterline = 0x4417U;

	  barrier();

	  // Probe the bridge to see if it responds.

	  if (!elmo_probe_bridge()) {
	    elmo_bridge_is_alive();
	  }

	}
	break;
      }
    case kElmoSetDebugMode:
      {
	debug = (int) arg;
	printk(KERN_INFO "elmo: debug level set to %d\n", debug);
	break;
      }
    case kElmoReserve:
      {
	break; /* not yet implemented, treat as a NO-OP */
      }
    case kElmoRelease:
      {
	break; /* not yet implemented, treat as a NO-OP */
      }
    case kElmoReportDiagMode:
      {
	elmo_host_data.diagMode = (int) arg;
	break;
      }
    case kElmoResetESS:
      {
	printk(KERN_INFO "elmo: rude reset of ESS\n");

	save_and_cli(flags);

	ELMO_BRIDGE->control |= ELMO_CONTROL_ESS_RST;
	barrier();
	ELMO_BRIDGE->control &= ~ELMO_CONTROL_ESS_RST;
        elmo_host_data.essWedged = 0;

	restore_flags(flags);
	
	break;
      }
    case kElmoResetFIFOs:
      {
	printk(KERN_INFO "elmo: flushing fifos\n");

	save_and_cli(flags);

        ELMO_BRIDGE->control |=  ELMO_CONTROL_LL_FIFO_RST;
	barrier();
        ELMO_BRIDGE->control &= ~ELMO_CONTROL_LL_FIFO_RST;
        barrier();
        ELMO_BRIDGE->control |=  ELMO_CONTROL_IDE_FIFO_RST;
        barrier();
        ELMO_BRIDGE->control &= ~ELMO_CONTROL_IDE_FIFO_RST;
        barrier();

	restore_flags(flags);
	
	break;
      }
    case kElmoDirectIO:
      {
          if (NULL == (void *)arg) {
              return -EINVAL;
          }

          struct elmo_request *cur_req = NULL;
          cur_req = kmalloc(sizeof(struct elmo_request), GFP_KERNEL);
          if (NULL == cur_req) {
              return -ENOMEM;
          }

          Sgat_dio_hdr_t *user_dio = (Sgat_dio_hdr_t *)arg;

          result = elmo_initiate_bulk_transfer(cur_req, user_dio);
          if (result < 0 ) {
              return result;
          }
  
          return elmo_complete_bulk_transfer();
      }
    case kElmoDirectIOAsync:
      {
          if (NULL == (void *)arg) {
              return -EINVAL;
          }

          struct elmo_request *cur_req = NULL;
          cur_req = kmalloc(sizeof(struct elmo_request), GFP_KERNEL);
          if (NULL == cur_req) {
              return -ENOMEM;
          }

          Sgat_dio_hdr_t *user_dio = (Sgat_dio_hdr_t *)arg;

          return elmo_initiate_bulk_transfer(cur_req, user_dio);
      }
    case kElmoCompleteDirectIOAsync:
      {
          return elmo_complete_bulk_transfer();
      }
    case kElmoCancelDirectIOAsync:
      {
          if (0 != elmo_host_data.bulkIOCompleted) {
              (void)elmo_abort_bulk_worker();
              result = elmo_complete_bulk_transfer();
          }
          return result;
      }
    case kElmoHasTei:
      {
          unsigned int loaded = elmo_host_data.pldLoaded;
          result = copy_to_user((void *)arg, &loaded, sizeof(unsigned int));
          break;
      }
    case kElmoGetEbiClkSpeed:
      {
          unsigned long ebi_speed = elmo_get_ebi_clock_speed();
          return copy_to_user((void *)arg, &ebi_speed, sizeof(unsigned long));
      }
    case kElmoDisableWatchDog:
      {
          noesswatchdog = 1;
          break;
      }
    case kElmoEnableWatchDog:
      {
          elmo_host_data.essWedged = 0;
          noesswatchdog = 0;
          break;
      }
    default:
      {
	result = -EINVAL;
      }
    }
  return result;
}

int elmo_ll_open (struct inode *i, struct file *filp)
{
  int type = MINOR(i->i_rdev);

  if (type > 0) {
    filp->f_op = &elmo_ll_deferred_fops;
  }

  return 0;

}

#ifdef MODULE

#include <linux/module.h>

#endif

EXPORT_SYMBOL(ELMOReadBridge);
EXPORT_SYMBOL(ELMOWriteBridge);
EXPORT_SYMBOL(ELMOWriteBridgeMasked);
